# generic-objects.py
#
# Copyright 2006 by Paul Emsley, The University of York
# Copyright 2006 by Bernhard Lohkamp
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import coot
import coot_utils

global probe_command
global reduce_command
global reduce_molecule_updates_current
global interactive_probe_is_OK_qm

# gets set the first time we run interactive_probe.  Takes values
# 'unset' (initial value) 'yes' and 'no')
#
interactive_probe_is_OK_qm = 'unset'

# set reduce  and probe command full path here if you wish
# (or have a look in group_settings!! These overwrite whatever is here), e.g.
# reduce_command = "/home/bernhard/bin/reduce"

# BL says:: just to be consistent with Paul's names
# map to scheme names:
# deftexi generic_object_is_displayed_qm
generic_object_is_displayed_qm = coot.generic_object_is_displayed_p

# map to scheme names:
# deftexi is_closed_generic_object_qm 
is_closed_generic_object_qm = coot.is_closed_generic_object_p

# return a new generic object number for the given object obj-name.
# If there is no such object with name obj-name, then create a new
# one.  (Maybe you want the generic object to be cleaned if it exists
# before returning, this function does not do that).
# 
def generic_object_with_name(obj_name):

  t = coot.generic_object_index(obj_name)
  if (t == -1):
    t = coot.new_generic_object_number(obj_name)
  return t


# display a GUI for generic objects
#
def generic_objects_gui():

    print(":::::::::::::::::: generic_objects_gui() needs a redo")
    # this is old. just call the build-in one.

    return

    pygtk_flag = False
    try:
      import pygtk
      pygtk.require("2.0")
      import gtk, pango
      pygtk_flag = True
    except:
      print("Fail")

    if (pygtk_flag and coot_utils.using_gui()):
      # Now we run the gui
      def delete_event(*args):
        # BL says: first we shall close the generic objects
        # or not
        #for generic_object_number in range(n_objects):
        # coot.set_display_generic_object(generic_object_number, 0)
        gen_window.destroy()
        return False


      def check_button_callback(widget, generic_object_number):
        button_state = widget.get_active()
        object_state = generic_object_is_displayed_qm(generic_object_number)
        if ((button_state == True) and (object_state == 0)):
          coot.set_display_generic_object(generic_object_number,1)
        if ((button_state == False) and (object_state == 1)):
          coot.set_display_generic_object(generic_object_number,0)

      def all_check_button_callback(widget):
        show_all = widget.get_active()
        for check_button in open_generic_objects:
          if (show_all):
            check_button.set_active(True)
          else:
            check_button.set_active(False)

      n_objects = coot.number_of_generic_objects()

      if (n_objects > 0):
        gen_window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        gen_window.set_title("Generic objects")
        vbox = gtk.VBox(False, 0)
        open_generic_objects = []
        no_active_generic_objects = 0

        for generic_object_number in range(n_objects):

          print("INFO:: generic object attributes: ", \
                generic_object_number, \
                generic_object_name(generic_object_number), \
                is_closed_generic_object_qm(generic_object_number))

          if (is_closed_generic_object_qm(generic_object_number) == 0):
            name = generic_object_name(generic_object_number)

            if (name):
              label = str(generic_object_number) + "  " + name
              frame = gtk.Frame(label=None)
              check_button = gtk.CheckButton(label)

                # this callback gets called by the
                # gtk-toggle-button-set-active just below,
                # which is why we need the state and active
                # test.
              check_button.connect("toggled", 
                     check_button_callback, generic_object_number)
              current_state = generic_object_is_displayed_qm(generic_object_number)
              if (current_state == 1):
                check_button.set_active(True)
                no_active_generic_objects += 1

              vbox.add(frame)
              frame.add(check_button)
              frame.show()
              check_button.show()
              open_generic_objects.append(check_button)

        if (len(open_generic_objects) > 1):
          hsep = gtk.HSeparator()
          label = "Show/hide all"
          frame = gtk.Frame(label=None)
          check_button = gtk.CheckButton(label)
          check_button.connect("toggled", all_check_button_callback)
          if (len(open_generic_objects) == no_active_generic_objects):
            check_button.set_active(True)
          vbox.add(hsep)
          hsep.show()
          vbox.add(frame)
          frame.add(check_button)
          frame.show()
          check_button.show()

        gen_window.connect("delete_event", delete_event)

        vbox.show()
        gen_window.add(vbox)
        gen_window.set_border_width(10)
        gen_window.show()


# return status
#
def reduce_on_pdb_file(imol, pdb_in, pdb_out):
    return reduce_on_pdb_file_generic(imol, "build", pdb_in, pdb_out)

# return status
#
def reduce_on_pdb_file_no_flip(imol, pdb_in, pdb_out):
    r = reduce_on_pdb_file_generic(imol, "no-flip", pdb_in, pdb_out)
    return r

# return status
#
def reduce_on_pdb_file_generic(imol, no_flip_or_build, pdb_in, pdb_out):

  global reduce_command
  
  print("running reduce on", pdb_in)
  # could try coot_utils.find_exe too!
  if not coot_utils.command_in_path_qm(reduce_command):
    print("command for reduce %s is not found in path" %reduce_command)
  else:
    # need full path to find het dict
    full_reduce_command = coot_utils.find_exe(reduce_command, "PATH")

    nshl = coot.non_standard_residue_names_py(imol)
    print("DEBUG:: in reduce_on_pdb_file_generic(), nshl is ", nshl)
    ext = "-".join(nshl)
    reduce_het_dict_file_name = "coot-molprobity/reduce-het-dict-" + ext + ".txt"
    write_reduce_het_dict(imol, reduce_het_dict_file_name)

    dict_args = []
    if not set_reduce_het_dict():
      # now should use the and build the het dic!?
      dict_args = ["-DB", reduce_het_dict_file_name]

    mode = "-build"
    if no_flip_or_build == "no-flip":
        mode = "-NOFLIP"

    print("======= reduce_on_pdb_file: command %s args %s with pdb_out: %s" \
          %(reduce_command, [mode, pdb_in] + dict_args, pdb_out))
    status = coot_utils.popen_command(reduce_command,
                           [mode, pdb_in] + dict_args,
                           [],
                           pdb_out)

    print("======== status", status)

    # 20130107 returning (status == 0) is problematic because sometimes/often reduce exits
    # with status 1 but it has worked OK!
    #
    # return status == 0
    return True


global old_pdb_style
old_pdb_style = False
  
reduce_molecule_updates_current = False

reduce_command = "molprobity.reduce"
probe_command = "molprobity.probe"

# run molprobity (well reduce and probe) to make generic objects (and
# display the generic objects gui)
#
def probe(imol):
  import os
  import coot_utils
  global reduce_command, probe_command
  global old_pdb_style
    
  if coot.is_valid_model_molecule(imol):

    if not (os.path.isfile(reduce_command)):
      reduce_command = coot_utils.find_exe("reduce", "PATH", "CBIN", "CCP4_BIN")
    # we need to check if probe_command is defined too
    if not(os.path.isfile(probe_command)):
      probe_command = coot_utils.find_exe("probe", "PATH", "CBIN", "CCP4_BIN")
    coot.make_directory_maybe("coot-molprobity")
    mol_pdb_file = "coot-molprobity/for-reduce.pdb"
    reduce_out_pdb_file = "coot-molprobity/reduced.pdb"
    reduce_het_dict_file_name = "coot-molprobity/reduce-het-dict.txt"
    coot.write_pdb_file(imol, mol_pdb_file)
    write_reduce_het_dict(imol, reduce_het_dict_file_name)
    if not reduce_command:
      # couldnt find reduce
      print("BL WARNING:: Could not locate the program reduce!! Please check if installed!")
    else:

      dict_args = []
      if not set_reduce_het_dict():
        dict_args = ["-DB", reduce_het_dict_file_name]

      if old_pdb_style:
        # old
        arg_list = ["-build", "-oldpdb", mol_pdb_file]
      else:
        # modern 
        arg_list = ["-build", mol_pdb_file]

      arg_list += dict_args

      print("BL INFO:: running reduce: %s %s and ouptut to: %s" \
            %(reduce_command , arg_list, reduce_out_pdb_file))
      print("BL INFO:: running reduce: REDUCE_HET_DICT env var:", os.getenv('REDUCE_HET_DICT'))
      
      reduce_status = coot_utils.popen_command(reduce_command,
                                    arg_list,
                                    [],
                                    reduce_out_pdb_file)
      
      # dont check for status as meaningless...
      if not probe_command:
        # couldnt find probe
        print("BL WARNING:: Could not locate the program probe!! Please check if installed!")
      else:
        probe_name_stub = coot_utils.strip_extension(coot_utils.strip_path(coot.molecule_name(imol)))
        probe_pdb_in = "coot-molprobity/" + probe_name_stub + "-with-H.pdb"
        probe_out = "coot-molprobity/probe-dots.out"

        prepare_file_for_probe(reduce_out_pdb_file, probe_pdb_in)

        probe_status = coot_utils.popen_command(probe_command,
                                     ["-u", "-mc", "ALL", probe_pdb_in],
                                     [],
                                     probe_out)

        if (probe_status):
          print("BL WARNING:: probe failed, cannot continue!")
        else:
          # by default, we don't want to click on the
          # imol-probe molecule (I think :-)
          recentre_status = coot.recentre_on_read_pdb()
          novalue = coot.set_recentre_on_read_pdb(0)
          if (reduce_molecule_updates_current):
            print("======= update molecule =======")
            imol_probe = coot.clear_and_update_model_molecule_from_file(imol, probe_pdb_in)
          else:
            print("======= read new pdb file =======")
            imol_probe = coot.read_pdb(probe_pdb_in)

          if recentre_status == 1:
            coot.set_recentre_on_read_pdb(1)

          # show the GUI for USER MODS
          if coot_utils.using_gui():
            coot_gui.user_mods_gui(imol_probe, reduce_out_pdb_file)

          # coot_utils.toggle_active_mol(imol_probe) let's not do
          # that actually.  I no longer think that the
          # new probe molecule should not be clickable
          # when it is initally displayed (that plus
          # there is some active/displayed logic problem
          # for the molecules, which means that after
          # several probes, the wrong molecule is
          # getting refined).

          coot.handle_read_draw_probe_dots_unformatted(probe_out, imol_probe, 2)
          # generic_objects_gui() old
          coot.graphics_draw()


# Write the connectivity for the non-standard (non-water) residues in
# the given molecule for which we have the dictionary.
#
# Don't return anything interesting.  
#
def write_reduce_het_dict(imol, reduce_het_dict_file_name):

  print("debug:: reduce_het_dict_file_name", reduce_het_dict_file_name)

  con_file_names = []
  for res_name in coot.non_standard_residue_names_py(imol):
      f_name = "coot-molprobity/conn-" + res_name + ".txt"
      status = coot.write_connectivity(res_name, f_name)
      if (status == 1):
          con_file_names.append(f_name)
  if con_file_names:
      # fin = open(reduce_het_dict_file_name, 'w')
      #for file_name in con_file_names:
      #    shutil.copyfileobj(open(file_name, 'rb'), fin)
      # fin.close()

      with open(reduce_het_dict_file_name, 'w') as out_file:
          for file_name in con_file_names:
              print("debug:: file_name", file_name)
              with open(file_name) as in_file:
                  out_file.write(in_file.read())

# Prepare file for probe, i.e. remove 'USER' from file
def prepare_file_for_probe(file_in, file_out):
    
    try:
      fin = open(file_in,'r')
    except IOError:
      print("BL WARNING:: Cannot read ", file_in)
    try:
      fout = open(file_out,'w')
    except IOError:
      print("BL WARNING:: Cannot write ", file_out)
    if (fin and fout):
       lines = fin.readlines()
       for line in lines:
           if (not "USER" in line):
               fout.write(line)
       fin.close()
       fout.close()

 
# run "probe" interactively, 
# which in the current implementation, means that this function 
# can run during a edit-chi angles manipulation, or after
# a real space refine zone.
# 
# Thus function presumes that there are 2 pdb files in the current 
# directory, one of which is the reference pdb file and the other
# is a pdb file containing the tmp/moving atom set.
# 
# The function takes arguments for the centre of the probe dots
# and the radius of the probe dots sphere.  The chain id and 
# residue number are also needed to pass as arguments to probe.
#
def interactive_probe(x_cen, y_cen, z_cen, radius, chain_id, res_no):

    import os, string
    global probe_command
    global interactive_probe_is_OK_qm

    probe_pdb_in_1 = "molprobity-tmp-reference-file.pdb"
    probe_pdb_in_2 = "molprobity-tmp-moving-file.pdb"
    probe_out = "coot-molprobity/molprobity-tmp-probe-dots.out"
    chain_str = ""
    if (not chain_id == ""):
      chain_str = "chain"
    atom_sel = "(file1 within " + str(radius) + " of " \
	       + str(x_cen) + ", " \
	       + str(y_cen) + ", " \
	       + str(z_cen) + ", " \
	       + "not water not (" + chain_str + string.lower(chain_id) + " " \
	       + str(res_no) + ")),file2"

    # no longer use std-bonds
    print("probe command", probe_command, \
          ["-mc", "-u", "-quiet", "-drop", "-both",
          atom_sel, "file2", probe_pdb_in_1, probe_pdb_in_2])

    # if unset, then set it.
    if (interactive_probe_is_OK_qm == 'unset'):
      if (coot_utils.command_in_path_qm(probe_command)):
        interactive_probe_is_OK_qm = 'yes'
      else:
        interactive_probe_is_OK_qm = 'no'
        
    if (interactive_probe_is_OK_qm == 'yes'):
       status = coot_utils.popen_command(probe_command,
			      ["-mc", "-u", "-quiet", "-drop", "-both",
			       atom_sel, "file2",
			       probe_pdb_in_1, probe_pdb_in_2],
			      [],
			      probe_out)

       # don't show the gui, so the imol is not needed/dummy.
       coot.handle_read_draw_probe_dots_unformatted(probe_out, 0, 0)
       coot.graphics_draw()

#
#
def get_probe_dots_from(pdb_file_name, point, radius):

  global probe_command
  global interactive_probe_is_OK_qm
  import os
  # if unset, then set it, try to make dir too
  if (interactive_probe_is_OK_qm == 'unset'):
    if (not coot_utils.command_in_path_qm(probe_command)):
      interactive_probe_is_OK_qm = 'no'
    else:
      dir_name = coot_utils.get_directory("coot-molprobity")
      if (dir_name):    # OK, we had it or made it
        interactive_probe_is_OK_qm = 'yes'
      else:
        interactive_probe_is_OK_qm = 'no'

  if (interactive_probe_is_OK_qm == 'yes'):
    probe_out = os.path.join("coot-molprobity", "molprobity-tmp-probe-dots.out")
    within_str = "(within " + \
                 str(radius) + \
                 " of " + \
                 str(point[0]) + \
                 ", " + \
                 str(point[1]) + \
                 ", " + \
                 str(point[2]) + \
                 ")"
    args = ["-mc", "-u", "-quiet", "-drop", "-stdbonds",
            "ALL",  # whole residues from sphere selection
                    # was needed to make this work
            # within_str  # problems with atom selection
            pdb_file_name]
    print("popen_comand on", probe_command, args)
    coot_utils.popen_command(probe_command, args, [], probe_out, False)
    coot.handle_read_draw_probe_dots_unformatted(probe_out, 0, 0)
    coot.graphics_draw()

# Update the generic objects probe dots from residues within radius
# of the screen centre.
# 
# Return nothing interesting.
#
def probe_local_sphere(imol, radius):

  # We need to select more atoms than we probe because if the atom
  # selection radius and the probe radius are the same, then
  # sometimes the middle atom of a bonded angle set is missing
  # (outside the sphere) and that leads to bad clashes.
  # There are also bad clashed at the edge when alt-confed atoms
  # are not selected but non-alt-confs are which leads to missing atoms
  # in a bond angle and therefore clashes.

  pt = coot_utils.rotation_centre()
  imol_new = coot.new_molecule_by_sphere_selection(imol, pt[0], pt[1], pt[2],
                                              radius, 0)
  coot.set_mol_displayed(imol_new, 0)
  set_mol_active   (imol_new, 0)
  pdb_name = "molprobity-tmp-reference-file.pdb"
  coot.make_directory_maybe("coot-molprobity")
  write_pdb_file_for_molprobity(imol_new, pdb_name)

  get_probe_dots_from(pdb_name, pt, radius)
  coot.close_molecule(imol_new)


def probe_local_sphere_active_atom(radius=5.0):
  
  with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                             aa_ins_code, aa_atom_name, aa_alt_conf]:
    probe_local_sphere(aa_imol, radius)

  
# add in the conn files by concatting.
#
def write_pdb_file_for_molprobity(imol, pdb_name):

  tmp_pdb_name = coot_utils.file_name_sans_extension(pdb_name) + "-tmp.pdb"

  coot.write_pdb_file(imol, tmp_pdb_name)

  # Let's add on the connectivity cards of the residues that
  # molprobity doesn't know about (which I presume are all
  # non-standard residues).  Cut out (filter) files that didn't
  # write properly.
  #
  conn_file_names = []
  for res_name in non_standard_residue_names(imol):
    f_name = os.path.join("coot-molprobity", "conn-" + res_name + ".txt")
    status = coot.write_connectivity(res_name, f_name)
    if (status == 1):
      conn_file_names.append(f_name)

  # now, add (append) each of the con-file-names to the end of
  # pdb-name
  if (not os.path.isfile(tmp_pdb_name)):
    print("ERROR:: tmp file name not found", tmp_pdb_name)
  else:
    conn_file_names.insert(0, tmp_pdb_name)
    fn_all = open(pdb_name, 'w')
    for conn_file in conn_file_names:
      if (os.path.isfile(conn_file)):
        cf = open(conn_file, 'r')
        data = cf.read()
        cf.close()
        fn_all.write(data)
    fn_all.close()

# not sure if the following ones should reside here?! Maybe rather in coot_gui.coot_gui?

# a toggle function for the main toolbar to switch on probe dots post refine
# and for chis/rotamers
#
def toggle_interactive_probe_dots(widget=None):
  """a toggle function for the main toolbar to switch on probe dots post refine
  and for chis/rotamers

  Keyword arguments:
  widget -- can be passed from the toolbutton
  
  """

  if widget:
    if widget.get_active():
      # the button is toggled on
      coot.set_do_probe_dots_on_rotamers_and_chis(1)
      coot.set_do_probe_dots_post_refine(1)
    else:
      coot.set_do_probe_dots_on_rotamers_and_chis(0)
      coot.set_do_probe_dots_post_refine(0)
  else:
    # no alternative for now (could just go by state and change back and forth)
    print("BL WARNING:: no widget")


def set_reduce_het_dict():
  """helper function to set env variable REDUCE_HET_DICT
  returns True if could be set (or already set), False if it couldnt be set.
  
  """

  global reduce_command

  # BL says: I think we should set REDUCE_HET_DICT
  # so let's set REDUCE_HET_DICT if not set already!
  # not sure if needed any more since we write a connectivity
  # file - let's see FIXME!!
  # if we have probe/reduce from ccp4 we should pick it up there!?
  # should be in $CCP4/share/reduce/

  red_het_dict = os.getenv('REDUCE_HET_DICT')
  red_het_file = "reduce_wwPDB_het_dict.txt"
  
  #negative path
#  if (not red_het_dict or
#      (red_het_dict and not os.path.isfile(red_het_dict))):

  # positive path (maybe should check that red_het_dict is a string?!)
  if (red_het_dict and
      (os.path.isfile(red_het_dict))):
    return True
  else:
    # not set and/or not found

    # first check if from ccp4 then it is in $CCP4/share/reduce/
    ccp4_dir = os.getenv('CCP4')  
    if (ccp4_dir and (os.path.normpath(ccp4_dir) in os.path.normpath(reduce_command))):
      # have ccp4 reduce
      red_het_dict = os.path.join(ccp4_dir, "share", "reduce",
                                 red_het_file)
      if os.path.isfile(red_het_dict):
        os.environ['REDUCE_HET_DICT'] = red_het_dict
        return True
      else:
        print("BL WARNING:: could neither find nor set REDUCE_HET_DICT !")
        return False
      
    else:
      # not from ccp4
      
      # we assume the dic is in same dir as reduce command
      full_reduce_command = coot_utils.find_exe(reduce_command, "PATH")
      dir, tmp = os.path.split(full_reduce_command)
      red_het_dict = os.path.join(dir, red_het_file)
      if (os.path.isfile(red_het_dict)):
        os.environ['REDUCE_HET_DICT'] = red_het_dict
        return True
      else:
        # before we give up check in share/coot
        # this is where the windows installer shall put it
        prefix_dir = os.getenv("COOT_PREFIX")
        if not prefix_dir:
          pkg_data_dir = coot.get_pkgdatadir_py()
        else:
          pkg_data_dir = os.path.join(prefix_dir, "share", "coot")
        red_het_file = os.path.join(pkg_data_dir, red_het_file)
        if (os.path.isfile(red_het_dict)):
          os.environ['REDUCE_HET_DICT'] = red_het_dict
          return True
        else:
          # finally give up
          print("BL WARNING:: could neither find nor set REDUCE_HET_DICT !")
          return False
      




import sys
import unittest
import coot
import coot_utils
import coot_testing_utils

import os
sys.path.append(".")
# more general!?! append python-test dir, i.e. directory this file resides in
import inspect
this_file=inspect.getsourcefile(lambda:0)
test_dir=os.path.split(this_file)[0]
sys.path.append(test_dir)
# sys.path.append("../../coot/python-tests")

from TestPdbMtzFunctions    import *
from TestShelxFunctions     import *
from TestLigandFunctions    import *
from TestRNAGhostsFunctions import *
from TestSSMFunctions       import *
from TestNCSFunctions       import *
from TestUtilsFunctions     import *
from TestInternalFunctions  import *

# class to write output of unittest into a 'memory file' (unittest_output)
# as well as to sys.stdout
class StreamIO:
        
    def __init__(self, src=sys.stderr, dst=sys.stdout):
        import io
        global unittest_output
        unittest_output = io.StringIO()
        self.src = src
        self.dst = dst
        self.extra = unittest_output

    def write(self, msg):
        #self.src.write(msg)
        self.extra.write(msg)
        self.dst.write(msg)

    def flush(self):
        self.src.flush()
        self.dst.flush()
        self.extra.flush()
        pass


suite = unittest.TestSuite()
test_list = [TestPdbMtzFunctions, TestShelxFunctions, TestLigandFunctions, TestRNAGhostsFunctions,
             TestSSMFunctions, TestNCSFunctions, TestUtilsFunctions, TestInternalFunctions]

# test_list = [TestRNAGhostsFunctions]

for test in test_list:
    suite.addTests(unittest.TestLoader().loadTestsFromTestCase(test))

log = StreamIO(sys.stderr, sys.stdout)

result = unittest.TextTestRunner(stream=log, verbosity=2).run(suite)

# Shouldnt we exit!?
# cheating?! We only exit Coot if we are not in graphics mode
if (coot.use_graphics_interface_state() == 0):
    if (result.wasSuccessful()):
        coot.coot_real_exit(0)
    else:
        coot.coot_real_exit(1)



import libBoostedMoleculesToTrianglesCXXClasses as CXXClasses
from libBoostedMoleculesToTrianglesCXXClasses import SceneSetup, Camera, CXXCoord_float, CompoundSelection, ColorScheme, MolecularRepresentationInstance, MyMolecule, Light, RendererGLSL, RendererGL, SolidColorRule
import sys

import sys
from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout, QPushButton, QMainWindow, QTextEdit
from PyQt5.QtOpenGL import QGLWidget
from PyQt5.QtGui import (QGuiApplication, QMatrix4x4, QOpenGLContext,
                         QOpenGLShader, QOpenGLShaderProgram, QSurfaceFormat, QWindow, QMouseEvent)
from PyQt5.QtCore import QPoint, Qt
from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtWidgets import QWidget, QGridLayout, QLineEdit, QLabel
import math

class WfWidget(QGLWidget):
    def __init__(self, parent = None):
        super(WfWidget, self).__init__(parent)
        self.prepareCXXClasses()
        self.setMouseTracking(True)
        self.oldMousePosition = QPoint(-1,-1)
        self.tested = False
    
    def mouseMoveEvent(self, event):
        newMousePosition = event.pos()
        deltaX = float(newMousePosition.x() - self.oldMousePosition.x())
        deltaY = float(newMousePosition.y() - self.oldMousePosition.y())
        magnitude = math.sqrt((deltaX*deltaX) + (deltaY*deltaY))
        if magnitude > 0:
            if event.buttons() & Qt.LeftButton :
                if event.modifiers() == Qt.NoModifier:
                    xComponent =  deltaY / magnitude
                    yComponent =  deltaX / magnitude
                    rotation = CXXCoord_float(magnitude/4., xComponent, yComponent, 0.)
                    self.camera.getSceneSetup().rotateBy(rotation)
                    self.updateGL()
            elif event.buttons() & Qt.RightButton :
                oldFovy = self.camera.getFovy()
                factor = 1. + (abs(deltaY)/100.)
                newFovy = oldFovy * factor
                if deltaY < 0: newFovy = oldFovy / factor
                self.camera.setFovy(newFovy)
                self.updateGL()
    
        self.oldMousePosition = newMousePosition
    
    def paintGL(self):
        self.makeCurrent()
        self.camera.renderWithRenderer(self.renderer)
    
    def resizeGL(self, w, h):
        pass

    def initializeGL(self):
        super(WfWidget,self).initializeGL()
        self.makeCurrent()
        print(self.context())
        self.makeCurrent()
        self.renderer = RendererGLSL()
        self.renderer.init()
        
        #self.renderer = RendererGL()
    
    def prepareCXXClasses(self):
        light = CXXClasses.Light.defaultLight()
        light.setDrawLight(False)
        sceneSetup = CXXClasses.SceneSetup.defaultSceneSetup()
        sceneSetup.addLight(light)
        self.camera = CXXClasses.Camera.defaultCamera()
        sceneSetup.addCamera(self.camera)
        self.camera.setSceneSetup(sceneSetup)

import sys
import code
from io import StringIO

class Interpreter(QObject, code.InteractiveConsole):
    output = pyqtSignal(str)

    def __init__(self, locals=None):
        QObject.__init__(self)
        code.InteractiveConsole.__init__(self, locals)
        self.out = StringIO()

    def write(self, data):
        self.output.emit(data)

    def runcode(self, codez):
        """
        Reimplementation to capture stdout and stderr
        """
        sys.stdout = self.out
        sys.stderr = self.out
        result = code.InteractiveConsole.runcode(self, codez)
        
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        sys.stdout.flush()
        sys.stderr.flush()
        self.output.emit(self.out.getvalue())

class MainWidget(QWidget):
    from PyQt5 import QtCore
    def __init__(self, *args, **kws):
        super(MainWidget,self).__init__(*args, **kws)
        layout = QGridLayout(self)
        self.display = QTextEdit()
        self.display.setReadOnly(True)
        layout.addWidget(self.display, 2,0, 1,2)
        self.display.setMaximumHeight(100);
        self.history = QTextEdit()
        self.history.setReadOnly(True)
        layout.addWidget(self.history, 2,2, 1,2)
        self.history.setMaximumHeight(100);
        self.prompt = QLabel(">>>")
        self.prompt.setMaximumWidth(100);
        layout.addWidget(self.prompt, 1,0)
        self.input = QLineEdit()
        self.input.installEventFilter(self)
        layout.addWidget(self.input, 1,1, 1, 3)
        self.wfWidget = WfWidget()
        self.wfWidget.setMinimumHeight(300);
        layout.addWidget(self.wfWidget, 0,0, 1,4)
        rotX=CXXCoord_float(1.,1.,0.,0.)
        myDict = {"camera":self.wfWidget.camera,"sceneSetup":self.wfWidget.camera.getSceneSetup()}
        combinedDict = {**globals(), **locals(), **myDict}
        self.interp = Interpreter(locals=combinedDict)
        self.input.returnPressed.connect(self.text_input)
        self.interp.output.connect(self.text_output)
        self.interp.runcode('''exec(open("QMGPy.py").read())''')
        self.inputLines = []
        self.inputLinesIndex = 0

    def eventFilter(self, object, event):
        from PyQt5 import QtCore
        if object == self.input:
            if event.type() == QtCore.QEvent.KeyPress:
                if event.key() == QtCore.Qt.Key_Up:
                    #print( "lineEdit -> Qt::Key_Up")
                    self.inputLinesIndex -= 1
                    if -1*self.inputLinesIndex <= len(self.inputLines):
                        replacementText = self.inputLines[self.inputLinesIndex]
                        self.input.setText(replacementText)
                    else:
                        self.inputLinesIndex = -1*len(self.inputLines)
                    return True
                elif event.key() == QtCore.Qt.Key_Down:
                    #print( "lineEdit -> Qt::Key_Down")
                    self.inputLinesIndex += 1
                    if self.inputLinesIndex <= -1:
                        replacementText = self.inputLines[self.inputLinesIndex]
                        self.input.setText(replacementText)
                    else:
                        self.inputLinesIndex = -1
                    return True
            return False
        return super(self, QWidget).eventFilter(obj, event)

    def text_input(self):
        text = self.input.text()
        self.inputLines.append(text)
        self.inputLinesIndex = 0
        self.history.append(">" + str(text))
        self.history.verticalScrollBar().setValue(self.history.verticalScrollBar().maximum());
        self.input.clear()
        
        if self.interp.push(str(text)):
            # More input required
            # Use sys.ps1 and sys.ps2
            self.prompt.setText("...")
        else:
            self.prompt.setText(">>>")
        self.wfWidget.update()


    def text_output(self, text):
        print(text)
        sys.stdout.flush()
        self.display.setPlainText(text)
        #self.display.append(text)
        self.display.verticalScrollBar().setValue(self.display.verticalScrollBar().maximum());


from PyQt5.QtWidgets import QTextEdit, QPushButton, QVBoxLayout

if __name__ == '__main__':
    app = QApplication(["CXXClasses"])
    mainWindow = MainWidget()
    mainWindow.show()
    sys.exit(app.exec_())



# populate_python_menus.py
# Copyright 2007, 2008 by Bernhard Lohkamp
# Copyright 2006, 2007, 2008 by The University of York
# Copyright 2015 by Medical Research Council
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,  51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA

# import pygtk, gtk, pango

import coot
import gi
gi.require_version('Gtk', '4.0')
from gi.repository import Gtk
import coot_gui_api
import time
import numbers
import coot_utils
import coot_gui
import fitting
import shelx_extensions
import gui_prosmart

# rename this file populate_python_menus.py

if coot_gui_api.main_menubar():

    def coot_menubar_menu_simple(s):
        menu = Gtk.Menu()
        menuitem = Gtk.MenuItem(s)
        menuitem.set_submenu(menu)
        main_menubar = coot_gui_api.main_menubar()
        main_menubar.append(menuitem)
        menuitem.show()
        return menu

    def coot_menubar_menu(menu_label_string):

        def get_menu_bar_label_dict():
            label_dict = {}
            for menu_child in coot_gui_api.main_menubar().get_children():
                label_dict[menu_child.get_children()[0].get_text()] = menu_child
            return label_dict

        try:
            menu_label_dict = get_menu_bar_label_dict()
            return menu_label_dict[menu_label_string].get_submenu()
        except KeyError as e:
            return coot_menubar_menu_simple(menu_label_string)

    def get_existing_submenu(menu, submenu_label):
        label_dict = {}
        for menu_child in menu.get_children():
            for c in menu_child.get_children():
                try:
                    t = c.get_text()
                    print("########### populate_python_menus get_existing_submenu get_text on c:", t)
                    if t == submenu_label:
                        return menu_child
                except KeyError as e:
                    pass
        return None


    # this should be added to the "Refine" module. It is not used here
    # (it doesn't belong in "Calculate")
    def add_refinement_options(submenu_refine):

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_refine,
            "Set Refinement Options...",
            lambda func: coot_gui.refinement_options_gui())

        submenu = Gtk.Menu()
        menuitem2 = Gtk.MenuItem("Peptide Restraints...")

        menuitem2.set_submenu(submenu)
        submenu_refine.append(menuitem2)
        menuitem2.show()

        def add_restr_func1():
            print('Planar Peptide Restraints added')
            coot.add_planar_peptide_restraints()

        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Add Planar Peptide Restraints",
            lambda func: add_restr_func1())


        def add_restr_func2():
            print('Planar Peptide Restraints removed')
            coot.remove_planar_peptide_restraints()

            coot_gui.add_simple_coot_menu_menuitem(
                submenu, "Remove Planar Peptide Restraints",
                lambda func: add_restr_func2())



    calculate_menu = coot_menubar_menu("Calculate")
    draw_menu      = coot_menubar_menu("Draw")
    edit_menu      = coot_menubar_menu("Edit")

    submenu_all_molecule   = Gtk.Menu()
    submenu_maps           = Gtk.Menu()
    submenu_modelling      = Gtk.Menu()
    submenu_ncs            = Gtk.Menu()
    submenu_views          = Gtk.Menu()
    submenu_representation = Gtk.Menu()
    submenu_pisa           = Gtk.Menu()
    submenu_modules        = Gtk.Menu()
    submenu_settings       = Gtk.Menu()

    # get rid of "all molecule" function for now - maybe for ever
    # calculate_all_molecule_menu_item = get_existing_submenu(calculate_menu, "All Molecule...")
    # calculate_all_molecule_menu_item.set_submenu(submenu_all_molecule)

    calculate_maps_menu_item = get_existing_submenu(calculate_menu, "Map Tools...")
    calculate_maps_menu_item.set_submenu(submenu_maps)

    calculate_model_menu_item = get_existing_submenu(calculate_menu, "Modelling...")
    calculate_model_menu_item.set_submenu(submenu_modelling)

    calculate_ncs_menu_item = get_existing_submenu(calculate_menu, "NCS Tools...")
    calculate_ncs_menu_item.set_submenu(submenu_ncs)

    calculate_pisa_menu_item = get_existing_submenu(calculate_menu, "PISA...")
    calculate_pisa_menu_item.set_submenu(submenu_pisa)

    calculate_modules_menu_item = get_existing_submenu(calculate_menu, "Modules...")
    calculate_modules_menu_item.set_submenu(submenu_modules)

    draw_representation_menu_item = get_existing_submenu(draw_menu, "Representation Tools...")
    draw_representation_menu_item.set_submenu(submenu_representation)

    edit_settings_menu_item = get_existing_submenu(edit_menu, "Settings...")
    edit_settings_menu_item.set_submenu(submenu_settings)

    draw_views_menu_item = Gtk.MenuItem("Views")
    draw_views_menu_item.set_submenu(submenu_views)
    draw_menu.append(draw_views_menu_item)
    draw_views_menu_item.show()


    # ---------------------------------------------------------------------
    #     Modules
    # ---------------------------------------------------------------------

    def add_module_user_defined_restraints():
        menu = coot_gui.coot_menubar_menu("Restraints")
        import user_define_restraints

    def add_calculate_modules_menu(submenu_modules):

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modules, "CCP4...",
            lambda func: coot_gui.add_module_ccp4())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modules, "SHELX...",
            lambda func: shelx_extensions.add_module_shelx())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modules, "User-defined Restraints...",
            lambda func: add_module_user_defined_restraints())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modules, "ProSMART",
            lambda func: gui_prosmart.add_module_prosmart())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modules, "Carbohydrate",
            lambda func: gui_add_linked_cho.add_module_carbohydrate_gui())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modules, "Cryo-EM",
            lambda func: coot_gui.add_module_cryo_em())

    # ---------------------------------------------------------------------
    #     Settings
    # ---------------------------------------------------------------------

    def add_edit_settings_menu(submenu_settings):

        submenu = Gtk.Menu()
        menuitem2 = Gtk.MenuItem("Rotate Translate Zone Mode...")

        menuitem2.set_submenu(submenu)
        submenu_settings.append(menuitem2)
        menuitem2.show()

        coot_gui.add_simple_coot_menu_menuitem(
            submenu, "Rotate About Fragment Centre",
            lambda func: coot.set_rotate_translate_zone_rotates_about_zone_centre(1))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu, "Rotate About Second Clicked Atom",
            lambda func: coot.set_rotate_translate_zone_rotates_about_zone_centre(0))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_settings,
            "Set Density Fit Graph Weight...",
            lambda func: coot_gui.generic_single_entry("set weight (smaller means apparently better fit)",
                                                       str("%.2f" %residue_density_fit_scale_factor()), "Set it",
                                                       lambda text: set_den_gra_func(text)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_settings,
            "Set Spin Speed",
            lambda func: coot_gui.generic_single_entry("Set Spin Speed (smaller is slower)",
                                                       str(coot.idle_function_rotate_angle()), "Set it",
                                                       lambda text: coot.set_idle_function_rotate_angle(float(text))))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_settings, "Nudge Centre...",
            lambda func: coot_gui.nudge_screen_centre_gui())


        def all_mol_symm_func():
            for imol in coot_utils.molecule_number_list():
                if coot_utils.valid_model_molecule_qm(imol):
                    coot.set_symmetry_whole_chain(imol, 1)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_settings, "All Molecules use \"Near Chains\" Symmetry",
            lambda func: [valid_model_molecule_qm(imol) and
                          coot.set_symmetry_whole_chain(imol, 1) for imol in coot_utils.molecule_number_list()])


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_settings, "Save Graphics Size and Positions",
            lambda func: coot.graphics_window_size_and_position_to_preferences())


        def save_dialog_func():
            coot.post_model_fit_refine_dialog()
            coot.post_go_to_atom_window()

            def delete_event(*args):
                window.destroy()
                return False

            def go_func(*args):
                save_dialog_positions_to_preferences_file()
                window.destroy()

            window = Gtk.Window(gtk.WINDOW_TOPLEVEL)
            label_text = "   When happy, press \"Save\" to save   \n" + "   dialog positions"
            label = Gtk.Label(label_text)
            h_sep = Gtk.HSeparator()
            cancel_button = Gtk.Button("  Cancel  ")
            go_button = Gtk.Button("  Save  ")
            vbox = Gtk.VBox(False, 4)
            hbox = Gtk.HBox(False, 4)

            hbox.pack_start(go_button,     False, False, 6)
            hbox.pack_start(cancel_button, False, False, 6)
            vbox.pack_start(label, False, False, 6)
            vbox.pack_start(h_sep, False, False, 6)
            vbox.pack_start(hbox,  False, False, 6)
            window.add(vbox)

            go_button.connect("clicked", go_func)
            cancel_button.connect("clicked", lambda w: window.destroy())

            window.show_all()

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_settings, "Save Dialog Positions...",
            lambda func: save_dialog_func())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_settings, "Key Bindings...", lambda func: coot_gui.key_bindings_gui())

        def install_and_show_key_bindings():
            coot_utils.file_to_preferences("template_key_bindings.py") # copy and evaluate
            coot_gui.key_bindings_gui()

            coot_gui.add_simple_coot_menu_menuitem(
                submenu_settings, "Python: Install Template Keybindings",
                lambda func: install_and_show_key_bindings())

        def quick_save_func(txt):
            try:
                n = int(txt)
                gobject.timeout_add(1000*n, quick_save)
            except:
                print("BL INFO:: could not add timer for auto save!")

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_settings, "Enable Quick-Save checkpointing...",
            lambda func: coot_gui.generic_single_entry("Checkpoint interval (seconds)",
                                                       "30",
                                                       " Start Auto-saving ",
                                                       lambda txt:
                                                       quick_save_func(txt)))

        def fit_protein_func1(imol):
            if coot.imol_refinement_map() == -1:
                coot.add_status_bar_text("oops. Must set a map to fit")
            else:
                global continue_multi_refine
                continue_multi_refine = True
                fitting.interruptible_fit_protein(imol, fitting.fit_protein_fit_function)

        def fit_protein_func2(imol):
            if coot.imol_refinement_map() == -1:
                coot.add_status_bar_text("oops. Must set a map to fit")
            else:
                global continue_multi_refine
                continue_multi_refine = True
                fitting.interruptible_fit_protein(imol, fitting.fit_protein_stepped_refine_function)

        def fit_protein_func3(imol):
            if coot.imol_refinement_map() == -1:
                coot.add_status_bar_text("oops. Must set a map to fit")
            else:
                global continue_multi_refine
                continue_multi_refine = True
                fitting.interruptible_fit_protein(imol, fitting.fit_protein_rama_fit_function)

        # --- add_edit_settings_menu() ends here

    def mask_map_func():
        f = ""
        molecule_list = coot_utils.molecule_number_list()
        if not molecule_list == []:
            for i in molecule_list:
                if coot.is_valid_map_molecule(molecule_list[i]):
                    print("%s is a valid map molecule" % molecule_list[i])
                    f = str(molecule_list[i])
                    break
            else:
                print("BL WARNING:: dunno what to do!? No map found")
                f = False
        return f

    def mask_map_func1(active_state):
        print("changed active_state to ", active_state)

    def mask_map_func2(imol, texts_list, invert_mask_qm):
        # map imol
        text_1 = texts_list[0]
        # atom selection
        text_2 = texts_list[1]
        # radius
        text_3 = texts_list[2]

        continue_qm = False
        try:
            n = int(text_1)
            continue_qm = True
        except:
            print(
                "BL WARNING:: input %s for Map molecule number is not an integer.\nBailing out" % n)

        if (continue_qm):
            if (invert_mask_qm):
                invert = 1
            else:
                invert = 0
            print("debug:: invert-mask? is", invert)
            if (text_3 != "default"):
                try:
                    new_radius = float(text_3)
                    coot.set_map_mask_atom_radius(new_radius)
                except:
                    print(
                        "BL WARNING:: could not set map mask radius to %s. It's not a number" % text_3)
            coot.mask_map_by_atom_selection(n, imol, text_2, invert)

    def mask_map_radius_func():
        rad = coot.map_mask_atom_radius()
        if (rad > 0):
            return str(rad)
        else:
            return "default"


    def add_calculate_maps_menu(submenu_maps):

        #---------------------------------------------------------------------
        #     Map Tools...
        #---------------------------------------------------------------------

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Mask Map by Atom Selection...",
            lambda func: coot_gui.molecule_chooser_gui("Define the molecule that has atoms to mask the map",
                                                       lambda imol: coot_gui.generic_multiple_entries_with_check_button(
                                                           [[" Map molecule number: ", mask_map_func],
                                                            [" Atom selection: ", "//A/1"],
                                                            ["Radius around atoms: ", mask_map_radius_func]],
                                                           [" Invert Masking? ", lambda active_state: mask_map_func1(active_state)],
                                                           "  Mask Map  ", lambda text_list, invert: mask_map_func2(imol, text_list, invert))))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Copy Map...",
            lambda func: coot_gui.map_molecule_chooser_gui("Map to Copy...",
                                                           lambda imol: coot.copy_molecule(imol)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Make a Smoother Copy...",
            lambda func: coot_gui.map_molecule_chooser_gui("Map Molecule to Smoothenize...",
                                                           lambda imol: coot.smooth_map(imol, 1.25)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Make a Very Smooth Copy...",
            lambda func: coot_gui.map_molecule_chooser_gui("Map Molecule to Smoothenize...",
                                                           lambda imol: coot.smooth_map(imol, 2.0)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Make a Difference Map...",
            lambda func: coot_gui.make_difference_map_gui())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Transform map by LSQ model fit...",
            lambda func: coot_gui.transform_map_using_lsq_matrix_gui())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Average Maps...",
            lambda func: coot_gui.average_map_gui())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Map Density Histogram...",
            lambda func: coot_gui.map_molecule_chooser_gui("Choose the map",
                                                           lambda imol: coot.map_histogram(imol)))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Brighten Maps",
            lambda func: coot_utils.brighten_maps())

        def set_diff_map_func(imol):
            print("setting map number %s to be a difference map" %imol)
            coot.set_map_is_difference_map(imol)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Set map is a difference map...",
            # do we need this wrapper?
            lambda func: coot_gui.map_molecule_chooser_gui("Which map should be considered a difference map?",
                                                           lambda imol: set_diff_map_func(imol)))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Another (contour) level...",
            lambda func: coot.another_level())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_maps,
            "Multi-chicken...",
            lambda func: coot_gui.map_molecule_chooser_gui("Choose a molecule for multiple contouring",
                                                           lambda imol: (coot.set_map_displayed(imol, 0), coot_utils.multi_chicken(imol))))




    def add_calculate_modelling_menu(submenu_maps):

        #---------------------------------------------------------------------
        #     Molecule functions/Modelling
        #
        #---------------------------------------------------------------------

        def add_hydrogens_with_coot_reduce():
            with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                       aa_ins_code, aa_atom_name, aa_alt_conf]:
                coot.coot_reduce(aa_imol)

        def add_hydrogens_refmac_func():
            with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                       aa_ins_code, aa_atom_name, aa_alt_conf]:
                coot_utils.add_hydrogens_using_refmac(aa_imol)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Add Hydrogens",
            lambda func: add_hydrogens_with_coot_reduce())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Add Hydrogens using Refmac",
            lambda func: add_hydrogens_refmac_func())

        coot_gui.add_simple_coot_menu_menuitem(
        submenu_modelling,
            "Add Other Solvent Molecules...",
            lambda func: coot_gui.solvent_ligands_gui())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Arrange Waters Around Protein...",
            lambda func: coot_gui.molecule_chooser_gui(
                "Arrange waters in molecule: ",
                lambda imol: coot.move_waters_to_around_protein(imol)))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Assign (force) HETATMs for this Residue",
            lambda func: coot_utils.using_active_atom(coot.hetify_residue,
                                                      "aa_imol", "aa_chain_id", "aa_res_no", "aa_ins_code"))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Assign HETATM to molecule...",
            lambda func: coot_gui.molecule_chooser_gui("Assign HETATMs as per PDB definition",
                                                       lambda imol: coot.assign_hetatms(imol)))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Atoms with Zero Occupancies...",
            lambda func: coot_gui.molecule_chooser_gui(
                "Which molecule to check for Atoms with zero occupancies?",
                lambda imol: coot_gui.zero_occ_atoms_gui(imol)))

        # --- D --------

        def delete_sidechains_for_active_chain():
            with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code, aa_atom_name, aa_alt_conf]:
                coot.delete_sidechains_for_chain(aa_imol, aa_chain_id)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Delete Side-chain Atoms for Active Chain",
            lambda x: delete_sidechains_for_active_chain())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Duplicate range (pick atoms)",
            lambda func: coot_gui.duplicate_range_by_atom_pick())


        # --- F --------

        def get_smiles_pdbe_func():
            with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code, aa_atom_name, aa_alt_conf]:
                comp_id = coot.residue_name(aa_imol, aa_chain_id, aa_res_no, aa_ins_code)
                coot_utils.get_SMILES_for_comp_id_from_pdbe(comp_id)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Fetch PDBe description for this ligand",
            lambda func: coot_utils.get_smiles_pdbe_func())

        def get_pdbe_ligand_func(comp_id):
            status = coot_utils.get_SMILES_for_comp_id_from_pdbe(comp_id)
            coot.get_monomer(comp_id)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Fetch PDBe Ligand Description",
            lambda func: coot_gui.generic_single_entry("Fetch PDBe Ligand Desciption for comp_id:",
                                                       "", " Fetch ", lambda comp_id: get_pdbe_ligand_func(comp_id)))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Fix Nomenclature Errors...",
            lambda func: coot_gui.molecule_chooser_gui("Fix Nomenclature Error in molecule:",
                                                       lambda imol: coot.fix_nomenclature_errors(imol)))


        # --- I --------

        def chiral_centre_inverter_func():
            with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                                  aa_ins_code, aa_atom_name, aa_alt_conf]:
                coot.invert_chiral_centre(aa_imol, aa_chain_id, aa_res_no, aa_ins_code, aa_atom_name)

        coot_gui.add_simple_coot_menu_menuitem(submenu_modelling,
                                               "Invert This Chiral Centre",
                                               lambda func:
                                               chiral_centre_inverter_func())


        # --- M ---

        def make_link_ext_func(*args):
            m_spec_1 = args[0]
            m_spec_2 = args[1]
            imol_1 = coot_utils.atom_spec_to_imol(m_spec_1)
            imol_2 = coot_utils.atom_spec_to_imol(m_spec_2)
            spec_1 = m_spec_1[2:]
            spec_2 = m_spec_2[2:]
            if not (imol_1 == imol_2):
                print("Mismatch molecules")
            else:
                coot.make_link_py(imol_1, spec_1, spec_2, "dummy", 0.1)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Make Link (click 2 atoms)...",
            lambda func:
            coot.user_defined_click_py(2, make_link_ext_func))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Merge Water Chains...",
            lambda func: coot_gui.molecule_chooser_gui("Merge Water Chains in molecule:",
                                                       lambda imol: coot_utils.merge_solvent_chains(imol)))


        def mon_dict_func(text):
            idealized = 0
            new_model = coot.get_monomer_from_dictionary(text, idealized)
            if not coot_utils.valid_model_molecule_qm(new_model):
                coot.get_monomer(text)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Monomer from Dictionary",
            lambda func:
            coot_gui.generic_single_entry("Pull coordinates from CIF dictionary for 3-letter-code:", "",
                                          " Get Coords ",
                                          lambda text: mon_dict_func(text)))


        # -- N --

        def new_mol_sphere_func1(imol, text):
            try:
                radius = float(text)
                args = [imol] + coot_utils.rotation_centre() + [radius, 0]
                coot.new_molecule_by_sphere_selection(*args)
            except:
                print("WARNING:: no valid radius", text)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "New Molecule by Sphere...",
            lambda func: coot_gui.generic_chooser_and_entry(
                "Choose a molecule from which to select a sphere of atoms:",
                "Radius:", "10.0",
                lambda imol, text: new_mol_sphere_func1(imol, text)))

        def new_mol_sym_func1(imol, text):
            from types import ListType
            pre_shift = origin_pre_shift(imol)
            if (type(pre_shift) is not ListType):
                print("bad pre-shift aborting")
            else:
                coot.new_molecule_by_symop(imol, text,
                                           pre_shift[0],
                                           pre_shift[1],
                                           pre_shift[2])

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "New Molecule from Symmetry Op...",
            lambda func: coot_gui.generic_chooser_and_entry(
                "Molecule from which to generate a symmetry copy",
                "SymOp", "X,Y,Z",
                lambda imol, text: new_mol_sym_func1(imol, text)))


        # --- P ---

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Phosphorylate this residue",
            lambda func: coot_utils.phosphorylate_active_residue())


        # ---- R ---------

        # --- Ren ---

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Rename Residue...",
            lambda func: coot_gui.rename_residue_gui())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Renumber Waters...",
            lambda func: coot_gui.molecule_chooser_gui(
                "Renumber waters of which molecule?",
                lambda imol: coot.renumber_waters(imol)))

        # --- Reo ---

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Reorder Chains...",
            lambda func: coot_gui.molecule_chooser_gui("Sort Chain IDs in molecule:",
                                                       lambda imol: coot.sort_chains(imol))) # an internal function

        # --- Res ---

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Residue Type Selection...",
            lambda func: coot_gui.generic_chooser_and_entry("Choose a molecule to select residues from: ","Residue Type:","",
                                                            lambda imol, text: (coot.new_molecule_by_residue_type_selection(imol, text),
                                                                                coot.update_go_to_atom_window_on_new_mol())))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Residues with Alt Confs...",
            lambda func: coot_gui.molecule_chooser_gui(
                "Which molecule to check for Alt Confs?",
                lambda imol: coot_gui.alt_confs_gui(imol)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Residues with Cis Peptide Bonds...",
            lambda func: coot_gui.molecule_chooser_gui("Choose a molecule for checking for Cis Peptides",
                                                       lambda imol: coot_gui.cis_peptides_gui(imol)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Residues with Missing Atoms...",
            lambda func: coot_gui.molecule_chooser_gui(
                "Which molecule to check for Missing Atoms?",
                lambda imol: coot_gui.missing_atoms_gui(imol)))

        # --- Rig ---

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Rigid Body Fit Residue Ranges...",
            lambda func:
            coot_gui.residue_range_gui(lambda imol, ls: coot.rigid_body_refine_by_residue_ranges_py(imol, ls),
                                       "Rigid Body Refine",
                                       "  Fit  "))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Rigid Body Fit Molecule...",
            lambda func: coot_gui.molecule_chooser_gui("Rigid Body Fit Molecule",
                                                       lambda imol: coot.rigid_body_refine_by_atom_selection(imol, "//")))

        # ---- S --------

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Superpose ligands",
            lambda func: coot_gui.superpose_ligand_gui())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "Symm Shift Reference Chain Here",
            lambda func: coot.move_reference_chain_to_symm_chain_position())

        # ---- W ---------

        def whats_this():
            central_residue = coot.active_residue_py()
            res_name = coot.residue_name(*central_residue[0:4])
            mol_no = central_residue[0]
            n = coot.comp_id_to_name_py(res_name)
            s = "(mol. no: " + str(mol_no) + ")  " + \
                res_name  + ":  " + \
                n if isinstance(n, str) else " <no-name-found>"
            coot.add_status_bar_text(s)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_modelling,
            "What's this?",
            lambda func: whats_this())


    def add_calculate_ncs_menu(submenu_ncs):

    #---------------------------------------------------------------------
    #     NCS functions
    #
    #---------------------------------------------------------------------

        def copy_ncs_range_func(imol, chain_id, text1, text2):
            try:
                r1 = int(text1)
                r2 = int(text2)
                if (chain_id == ncs.ncs_master_chain_id(imol)):
                    coot.copy_residue_range_from_ncs_master_to_others(imol, chain_id, r1, r2)
                else:
                    # different given master to current master
                    # ask what to do.
                    txt = "Current master chain is %s, but you asked to copy from %s.\n" \
                        %(ncs_master_chain_id(imol), chain_id)
                    txt += "Apply this master change?\n\n"
                    txt += "N.B. if no, then nothing is copied."
                    r = coot_gui.yes_no_dialog(txt, "Change Master")
                    if r:
                        coot.ncs_control_change_ncs_master_to_chain_id(imol, chain_id)
                        coot.copy_residue_range_from_ncs_master_to_others(imol, chain_id, r1, r2)
                        ## could change master back?!
                    else:
                        coot.info_dialog("Master chain was not changed and copy not applied.")
            except:
                print("WARNING:: no valid number input")

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_ncs,
            "Copy NCS Residue Range...",
            lambda func: coot_gui.generic_chooser_and_entry("Apply NCS Range from Master",
                                                            "Master Chain ID",
                                                            coot_utils.get_first_ncs_master_chain(),  # returns "" on fail
                                                            lambda imol, chain_id: coot_gui.generic_double_entry("Start of Residue Number Range",
                                                                                                                 "End of Residue Number Range",
                                                                                                                 "", "", False, False,
                                                                                                                 "Apply NCS Residue Range",
                                                                                                                 lambda text1, text2: copy_ncs_range_func(imol, chain_id, text1, text2))))


        def copy_ncs_chain_func(imol, chain_id):
            ncs_chains = ncs_chain_ids(imol)
            if (ncs_chains):
                # maybe this could be a function to avoid repetition.
                if (chain_id == ncs.ncs_master_chain_id(imol)):
                    # hunkey dorey
                    coot.copy_from_ncs_master_to_others(imol, chain_id)
                else:
                    # different given master to current master
                    # ask what to do.
                    txt = "Current master chain is %s, but you asked to copy from %s.\n" \
                        %(ncs_master_chain_id(imol), chain_id)
                    txt += "Apply this master change?\n\n"
                    txt += "N.B. if no, then nothing is copied."
                    r = coot_gui.yes_no_dialog(txt, "Change Master")
                    if r:
                        coot.ncs_control_change_ncs_master_to_chain_id(imol, chain_id)
                        coot.copy_from_ncs_master_to_others(imol, chain_id)
                        ## could change master back?!
                    else:
                        coot.info_dialog("Master chain was not changed and copy not applied.")
            else:
                s = "You need to define NCS operators for molecule " + str(imol)
                coot.info_dialog(s)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_ncs,
            "Copy NCS Chain...",
            lambda func: coot_gui.generic_chooser_and_entry("Apply NCS edits from NCS Master Chain to Other Chains",
                                                            "Master Chain ID",
                                                            coot_utils.get_first_ncs_master_chain(),  # can return  "".
                                                            lambda imol, chain_id: copy_ncs_chain_func(imol, chain_id)))


        def ncs_ghost_res_range_func(imol):

            def handle_go_function(resno_1_text, resno_2_text):
                cont = False
                try:
                    resno_1 = int(resno_1_text)
                    resno_2 = int(resno_2_text)
                    cont = True
                except:
                    print("WARNING:: input residue numbers have to be integers")
                if cont:
                    ghost_ncs_chain_ids = ncs_chain_ids(imol)
                    if (type(ghost_ncs_chain_ids) is ListType):
                        # because we can have hetero-NCS,
                        # but we ignore NCS other that
                        # that of the first type.
                        ghost_chain_list = ghost_ncs_chain_ids[0]
                        ncs.manual_ncs_ghosts(imol, resno_1, resno_2, ghost_chain_list)

            from types import ListType
            label_1 = "Start ResNumber"
            label_2 = "End ResNumber"
            entry_1_default_text = "10"
            entry_2_default_text = "20"
            go_button_label = "Regenerate Ghosts"

            coot_gui.generic_double_entry(label_1, label_2, entry_1_default_text,
                                          entry_2_default_text, False, False,
                                          go_button_label, handle_go_function)


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_ncs,
            "NCS Ghosts by Residue Range...",
            lambda func: coot_gui.molecule_chooser_gui("Make local NCS ghosts for molecule:",
                                                       lambda imol: ncs_ghost_res_range_func(imol)))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_ncs,
            "Update NCS Ghosts using Local Match",
            lambda func: ncs.update_ncs_ghosts_by_local_sphere())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_ncs,
            "NCS Jumping...",
            lambda func: coot_gui.ncs_jumping_gui())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_ncs,
            "NCS ligands...",
            lambda func: coot_gui.ncs_ligand_gui())

        submenu = Gtk.Menu()
        menuitem2 = Gtk.MenuItem("NCS matrix type...")

        menuitem2.set_submenu(submenu)
        submenu_ncs.append(menuitem2)
        menuitem2.show()

        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Accurate (SSM)",
            lambda func: coot.set_ncs_matrix_type(0))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Fast (LSQ)",
            lambda func: coot.set_ncs_matrix_type(1))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Extra Fast (LSQ, only every 2nd CA)",
            lambda func: coot.set_ncs_matrix_type(2))




    def add_draw_representations_menu(submenu_views):

        # ---------------------------------------------------------------------
        #     Views/Representations
        # ---------------------------------------------------------------------
        #

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Undo Symmetry View",
            lambda func: coot.undo_symmetry_view())


        def make_ball_n_stick_func(imol, text):
            bns_handle = coot.make_ball_and_stick(imol, text, 0.18, 0.3, 1)
            print("handle: ", bns_handle)


        global default_ball_and_stick_selection    # maybe should be at the top of the file

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Ball & Stick...",
            lambda func: coot_gui.generic_chooser_and_entry("Ball & Stick",
                                                            "Atom Selection:",
                                                            default_ball_and_stick_selection,
                                                            lambda imol, text: make_ball_n_stick_func(imol, text)))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Add Balls to Simple Sticks",
            lambda func: [set_draw_stick_mode_atoms(imol, 1) for imol in coot_utils.molecule_number_list()])


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Simple Sticks (No Balls)",
            lambda func: [set_draw_stick_mode_atoms(imol, 0) for imol in coot_utils.molecule_number_list()])


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Clear Ball & Stick...",
            lambda func: coot_gui.molecule_chooser_gui(
                "Choose a molecule from which to clear Ball&Stick objects",
                lambda imol: coot.clear_ball_and_stick(imol)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Electrostatic Surface...",
            lambda func: coot_gui.molecule_chooser_gui(
                "Choose a molecule to represent as a surface..." + \
                "\n" + \
                "Can be SLOW",
                lambda imol: coot.do_surface(imol, 1)))  # shall we switch on the light too?!


        def surface_func1(clipped = 0):
            active_atom = active_residue()
            aa_imol      = active_atom[0]
            aa_chain_id  = active_atom[1]
            aa_res_no    = active_atom[2]
            aa_ins_code  = active_atom[3]
            aa_atom_name = active_atom[4]
            aa_alt_conf  = active_atom[5]
            central_residue = active_residue()
            residues = residues_near_residue(aa_imol, central_residue[1:4], 6.0)
            # no waters in surface, thanks.
            # but what if they have different names?! (only HOH so far)
            filtered_residues = []
            for res in residues:
                if (coot.residue_name(aa_imol, *res) != "HOH"):
                    filtered_residues.append(res)
            imol_copy = coot.copy_molecule(aa_imol)
            # delete the interesting residue from the copy (so that
            # it is not surfaced).
            coot.delete_residue(imol_copy, aa_chain_id, aa_res_no, aa_ins_code)
            if clipped:
                coot.do_clipped_surface(imol_copy, filtered_residues)
            else:
                coot.do_surface(imol_copy, 1)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Clipped Surface Here (This Residue)",
            lambda func:
            surface_func1(1))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Full Surface Around Here (This Residue)",
            lambda func:
            surface_func1())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Un-Surface...",
            lambda func: coot_gui.molecule_chooser_gui(
                "Choose a molecule to represent conventionally...",
                lambda imol: coot.do_surface(imol, 0)))

        def hilight_site_func():
            active_atom = active_residue()
            if (active_atom):
                imol = active_atom[0]
                centre_residue_spec = [active_atom[1],
                                       active_atom[2],
                                       active_atom[3]]
                coot_utils.hilight_binding_site(imol, centre_residue_spec, 230,4)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Highlight Interesting Site (here)...",
            lambda func: hilight_site_func())


        def make_dot_surf_func(imol,text):
            # I think a single colour is better than colour by atom
            coot.set_dots_colour(imol, 0.5, 0.5, 0.5)
            coot.dots_handle = dots(imol, text, text, 2, 1)
            print("dots handle: ", dots_handle)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Dotted Surface...",
            lambda func: coot_gui.generic_chooser_and_entry("Surface for molecule",
                                                            "Atom Selection:", "//A/1-2",
                                                            lambda imol, text: make_dot_surf_func(imol, text)))


        def clear_dot_surf_func(imol,text):
            try:
                n = int(text)
                coot.clear_dots(imol,n)
            except:
              print("BL WARNING:: dots handle number shall be an integer!!")

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Clear Surface Dots...",
            lambda func: coot_gui.generic_chooser_and_entry("Molecule with Dotted Surface",
                                                            "Dots Handle Number:", "0",
                                                            lambda imol, text: clear_dot_surf_func(imol, text)))


        def limit_model_disp_func(text):
            try:
                f = float(text)
                if f < 0.1:
                    coot.set_model_display_radius(0, 10)
                else:
                    coot.set_model_display_radius(1, f)
            except:
                coot.set_model_display_radius(0, 10)

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Limit Model Display Radius...",
            lambda func: coot_gui.generic_single_entry("Display Radius Limit (0 for \'no limit\') ",
                                                       #  "15.0" ;; maybe this should be the map radius
                                                       # BL says:: I think it should be the current one
                                                       str(coot.get_map_radius()),
                                                       "Set: ",
                                                       lambda text: limit_model_disp_func(text)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "HOLE...",
            lambda func: test_hole.hole_ify())

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_representation,
            "Label All CAs...",
            lambda func: coot_gui.molecule_chooser_gui("Choose a molecule to label",
                                                       lambda imol: coot_utils.label_all_CAs(imol)))



        #---------------------------------------------------------------------
        #     3D annotations
        #---------------------------------------------------------------------

        submenu = Gtk.Menu()
        menuitem2 = Gtk.MenuItem("3D Annotations...")

        menuitem2.set_submenu(submenu)
        submenu_representation.append(menuitem2)
        menuitem2.show()

        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Annotate position...",
            lambda func: coot_gui.generic_single_entry("Annotation: ", "",
                                                       "Make Annotation",
                                                       lambda txt: coot_utils.add_annotation_here(txt)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Save Annotations...",
            lambda func: coot_gui.generic_single_entry("Save Annotations",
                                                       "coot_annotations.py",
                                                       " Save ",
                                                       lambda file_name: coot_utils.save_annotations(file_name)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Load Annotations...",
            lambda func: coot_gui.generic_single_entry("Load Annotations",
                                                       "coot_annotations.py",
                                                       " Load ",
                                                       lambda file_name: coot_utils.load_annotations(file_name)))


        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Remove annotation here",
            lambda func: coot_utils.remove_annotation_here())


        coot_gui.add_simple_coot_menu_menuitem(
            submenu,
            "Remove annotation near click",
            lambda func: coot_utils.remove_annotation_at_click())



    def add_draw_views_menu(submenu_views):

           coot_gui.add_simple_coot_menu_menuitem(
               submenu_views,
               "Add View...",
               lambda func: coot_gui.view_saver_gui())

           # BL says:: maybe check if number at some point
           coot_gui.add_simple_coot_menu_menuitem(
               submenu_views,
               "Add a Spin View...",
               lambda func: coot_gui.generic_double_entry("Number of Steps", 
                                                          "Number of Degrees (total)", "3600", "360", 
                                                          False, False,  #check button text and callback
                                                          "  Add Spin  ",
                                                          lambda text_1, text_2: coot.add_spin_view("Spin", int(text_1), float(text_2))))

           coot_gui.add_simple_coot_menu_menuitem(
               submenu_views,
               "Views Panel...",
               lambda func: coot_gui.views_panel_gui())

           coot_gui.add_simple_coot_menu_menuitem(
               submenu_views,
               "Play Views",
               lambda func: list(map(eval,["go_to_first_view(1)",
                                           "time.sleep(1)", "play_views()"])))

           # BL says:: maybe check if number at some point
           coot_gui.add_simple_coot_menu_menuitem(
               submenu_views, "Set Views Play Speed...",
               lambda func: coot_gui.generic_single_entry("Set Views Play Speed",
                                                          str(coot.views_play_speed()), "  Set it  ",
                                                          lambda text: coot.set_views_play_speed(float(text))))


           coot_gui.add_simple_coot_menu_menuitem(
               submenu_views, "Save Views...",
               lambda func: coot_gui.generic_single_entry("Save Views",
                                                          "coot-views.py", " Save ",
                                                          lambda txt: coot.save_views(txt)))


    def add_calculate_pisa_menu(submenu_pisa):
        # ---------------------------------------------------------------------
        #     PISA Interface and Assemblies
        # ---------------------------------------------------------------------

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_pisa, "PISA Assemblies...",
            lambda func:
            coot_gui.molecule_chooser_gui("Choose molecule for PISA assembly analysis",
                                          lambda imol:
                                          parse_pisa_xml.pisa_assemblies(imol)))

        coot_gui.add_simple_coot_menu_menuitem(
            submenu_pisa, "PISA Interfaces...",
            lambda func:
            coot_gui.molecule_chooser_gui("Choose molecule for PISA interface analysis",
                                          lambda imol:
                                          parse_pisa_xml.pisa_interfaces(imol)))



    #---------------------------------------------------------------------
    #  Run those functions to populate the menus
    #---------------------------------------------------------------------

    add_calculate_modules_menu(submenu_modules)
    add_calculate_maps_menu(submenu_maps)
    add_calculate_modelling_menu(submenu_modelling)
    add_calculate_ncs_menu(submenu_ncs)
    add_edit_settings_menu(submenu_settings)
    add_draw_representations_menu(submenu_representation)
    add_draw_views_menu(submenu_views)
    add_calculate_pisa_menu(submenu_pisa)



#!/usr/bin/env python
"""A log function that returns negative infinity for log(0) or log(x<0)"""

# Copyright 2010 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

from math import log

negInf = float("-infinity")   #a stand-in for negative infinity
#negInf = -999999999          #use this if float("-infinity") doesn't work (i.e. on Windows with Pythons earlier than 2.5)


def ln(a):
    """Calculate the log without raising an error
    
    ARGUMENTS:
        a - the value to calculate the log of
    RETURNS:
        log(a) if it is defined
        negative infinity otherwise (when a <= 0)
    """
    
    ln = None
    
    try:
        ln = log(a)
    except (ValueError, OverflowError):
        ln = negInf
    
    return ln



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import copy
from itertools import tee
import logging
from rdkit import Chem
from rdkit.Chem.rdchem import BondType, BondStereo, BondDir


__author__ = 'Matt Swain'
__email__ = 'm.swain@me.com'
__license__ = 'MIT'

log = logging.getLogger('tautomer')

BONDMAP = {'-': BondType.SINGLE, '=': BondType.DOUBLE, '#': BondType.TRIPLE, ':': BondType.AROMATIC}
CHARGEMAP = {'+': 1, '0': 0, '-': -1}

tautomer_transforms = [
    {'name': '1,3 (thio)keto/enol f', 'smarts': '[CX4!H0][C]=[O,S,Se,Te;X1]'},
    {'name': '1,3 (thio)keto/enol r', 'smarts': '[O,S,Se,Te;X2!H0][C]=[C]'},
    {'name': '1,5 (thio)keto/enol f', 'smarts': '[CX4,NX3;!H0][C]=[C][CH0]=[O,S,Se,Te;X1]'},
    {'name': '1,5 (thio)keto/enol r', 'smarts': '[O,S,Se,Te;X2!H0][CH0]=,:[C][C]=,:[C,N]'},
    {'name': 'aliphatic imine f', 'smarts': '[CX4!H0][C]=[NX2]'},
    {'name': 'aliphatic imine r', 'smarts': '[NX3!H0][C]=[CX3]'},
    {'name': 'special imine f', 'smarts': '[N!H0][C]=[CX3R0]'},
    {'name': 'special imine r', 'smarts': '[CX4!H0][c]=,:[n]'},
    {'name': '1,3 aromatic heteroatom H shift f', 'smarts': '[#7!H0][#6R1]=[O,#7X2]'},
    {'name': '1,3 aromatic heteroatom H shift r', 'smarts': '[O,#7;!H0][#6R1]=,:[#7X2]'},
    {'name': '1,3 heteroatom H shift', 'smarts': '[#7,S,O,Se,Te;!H0][#7X2,#6,#15]=[#7,#16,#8,Se,Te]'},
    {'name': '1,5 aromatic heteroatom H shift', 'smarts': '[n,s,o;!H0]:[c,n]:[c]:[c,n]:[n,s,o;H0]'},
    {'name': '1,5 aromatic heteroatom H shift f', 'smarts': '[#7,#16,#8,Se,Te;!H0][#6,nX2]=,:[#6,nX2][#6,#7X2]=,:[#7X2,S,O,Se,Te]'},
    {'name': '1,5 aromatic heteroatom H shift r', 'smarts': '[#7,S,O,Se,Te;!H0][#6,#7X2]=,:[#6,nX2][#6,nX2]=,:[#7,#16,#8,Se,Te]'},
    {'name': '1,7 aromatic heteroatom H shift f', 'smarts': '[#7,#8,#16,Se,Te;!H0][#6,#7X2]=,:[#6,#7X2][#6,#7X2]=,:[#6][#6,#7X2]=,:[#7X2,S,O,Se,Te,CX3]'},
    {'name': '1,7 aromatic heteroatom H shift r', 'smarts': '[#7,S,O,Se,Te,CX4;!H0][#6,#7X2]=,:[#6][#6,#7X2]=,:[#6,#7X2][#6,#7X2]=,:[NX2,S,O,Se,Te]'},
    {'name': '1,9 aromatic heteroatom H shift f', 'smarts': '[#7,O;!H0][#6,#7X2]=,:[#6,#7X2][#6,#7X2]=,:[#6,#7X2][#6,#7X2]=,:[#6,#7X2][#6,#7X2]=,:[#7,O]'},
    {'name': '1,11 aromatic heteroatom H shift f', 'smarts': '[#7,O;!H0][#6,nX2]=,:[#6,nX2][#6,nX2]=,:[#6,nX2][#6,nX2]=,:[#6,nX2][#6,nX2]=,:[#6,nX2][#6,nX2]=,:[#7X2,O]'},
    {'name': 'furanone f', 'smarts': '[O,S,N;!H0][#6X3r5;$([#6][!#6])]=,:[#6X3r5]'},
    {'name': 'furanone r', 'smarts': '[#6r5!H0][#6X3r5;$([#6][!#6])]=[O,S,N]'},
    {'name': 'keten/ynol f', 'smarts': '[C!H0]=[C]=[O,S,Se,Te;X1]', 'bonds': '#-'},
    {'name': 'keten/ynol r', 'smarts': '[O,S,Se,Te;!H0X2][C]#[C]', 'bonds': '=='},
    {'name': 'ionic nitro/aci-nitro f', 'smarts': '[C!H0][N+;$([N][O-])]=[O]'},
    {'name': 'ionic nitro/aci-nitro r', 'smarts': '[O!H0][N+;$([N][O-])]=[C]'},
    {'name': 'oxim/nitroso f', 'smarts': '[O!H0][N]=[C]'},
    {'name': 'oxim/nitroso r', 'smarts': '[C!H0][N]=[O]'},
    {'name': 'oxim/nitroso via phenol f', 'smarts': '[O!H0][N]=[C][C]=[C][C]=[OH0]'},
    {'name': 'oxim/nitroso via phenol r', 'smarts': '[O!H0][c]:[c]:[c]:[c][N]=[OH0]'},
    {'name': 'cyano/iso-cyanic acid f', 'smarts': '[O!H0][C]#[N]', 'bonds': '=='},
    {'name': 'cyano/iso-cyanic acid r', 'smarts': '[N!H0]=[C]=[O]', 'bonds': '#-'},
    {'name': 'formamidinesulfinic acid f', 'smarts': '[O,N;!H0][C][S,Se,Te]=[O]', 'bonds': '=--'},
    {'name': 'formamidinesulfinic acid r', 'smarts': '[O!H0][S,Se,Te][C]=[O,N]', 'bonds': '=--'},
    {'name': 'isocyanide f', 'smarts': '[C-0!H0]#[N+0]', 'bonds': '#', 'charges': '-+'},
    {'name': 'isocyanide r', 'smarts': '[N+!H0]#[C-]', 'bonds': '#', 'charges': '-+'},
    {'name': 'phosphonic acid f', 'smarts': '[OH][PH0]', 'bonds': '='},
    {'name': 'phosphonic acid r', 'smarts': '[PH]=[O]', 'bonds': '-'}
]
for transform in tautomer_transforms:
    transform['smarts'] = Chem.MolFromSmarts(transform['smarts'].encode('utf8'))

tautomer_scores = [
    {'name': 'benzoquinone', 'smarts': '[#6]1([#6]=[#6][#6]([#6]=[#6]1)=,:[N,S,O])=,:[N,S,O]', 'score': 25},
    {'name': 'oxim', 'smarts': '[#6]=[N][OH]', 'score': 4},
    {'name': 'C=O', 'smarts': '[#6]=,:[#8]', 'score': 2},
    {'name': 'N=O', 'smarts': '[#7]=,:[#8]', 'score': 2},
    {'name': 'P=O', 'smarts': '[#15]=,:[#8]', 'score': 2},
    {'name': 'C=hetero', 'smarts': '[#6]=[!#1;!#6]', 'score': 1},
    {'name': 'methyl', 'smarts': '[CX4H3]', 'score': 1},
    {'name': 'guanidine terminal=N', 'smarts': '[#7][#6](=[NR0])[#7H0]', 'score': 1},
    {'name': 'guanidine endocyclic=N', 'smarts': '[#7;R][#6;R]([N])=[#7;R]', 'score': 2},
    {'name': 'aci-nitro', 'smarts': '[#6]=[N+]([O-])[OH]', 'score': -4},
]
for tscore in tautomer_scores:
    tscore['smarts'] = Chem.MolFromSmarts(tscore['smarts'])

def tautomer_score(mol):

    smiles = Chem.MolToSmiles(mol, isomericSmiles=True)
    log.debug('Tautomer: %s', smiles)
    score = 0
    # Add aromatic ring scores
    ssr = Chem.GetSymmSSSR(mol)
    for ring in ssr:
        btypes = {mol.GetBondBetweenAtoms(*pair).GetBondType() for pair in _pairwise(ring)}
        elements = {mol.GetAtomWithIdx(idx).GetAtomicNum() for idx in ring}
        if btypes == {BondType.AROMATIC}:
            log.debug('Score +100 (aromatic ring)')
            score += 100
            if elements == {6}:
                log.debug('Score +150 (carbocyclic aromatic ring)')
                score += 150
    # Add SMARTS scores
    for tscore in tautomer_scores:
        for match in mol.GetSubstructMatches(tscore['smarts']):
            log.debug('Score %+d (%s)', tscore['score'], tscore['name'])
            score += tscore['score']
    # Add (P,S,Se,Te)-H scores
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() in {15, 16, 34, 52}:
            hs = atom.GetTotalNumHs()
            if hs:
                log.debug('Score %+d (%s-H bonds)', -hs, atom.GetSymbol())
                score -= hs

    return score


def canonical_tautomer(mol, max_tautomers=1000):
    """Enumerate all possible tautomers and return a canonical tautomer based on a scoring system.

    :param mol: An RDKit Mol object.
    :param max_tautomers: The maximum number of tautomers to enumerate (limit to prevent combinatorial explosion)
    """
    tautomers = enumerate_tautomers(mol, max_tautomers)
    if len(tautomers) == 1:
        return tautomers[0]
    # Calculate score for each tautomer
    highest = None
    for t in tautomers:
        smiles = Chem.MolToSmiles(t, isomericSmiles=True)
        log.debug('Tautomer: %s', smiles)
        score = 0
        # Add aromatic ring scores
        ssr = Chem.GetSymmSSSR(t)
        for ring in ssr:
            btypes = {t.GetBondBetweenAtoms(*pair).GetBondType() for pair in _pairwise(ring)}
            elements = {t.GetAtomWithIdx(idx).GetAtomicNum() for idx in ring}
            if btypes == {BondType.AROMATIC}:
                log.debug('Score +100 (aromatic ring)')
                score += 100
                if elements == {6}:
                    log.debug('Score +150 (carbocyclic aromatic ring)')
                    score += 150
        # Add SMARTS scores
        for tscore in tautomer_scores:
            for match in t.GetSubstructMatches(tscore['smarts']):
                log.debug('Score %+d (%s)', tscore['score'], tscore['name'])
                score += tscore['score']
        # Add (P,S,Se,Te)-H scores
        for atom in t.GetAtoms():
            if atom.GetAtomicNum() in {15, 16, 34, 52}:
                hs = atom.GetTotalNumHs()
                if hs:
                    log.debug('Score %+d (%s-H bonds)', -hs, atom.GetSymbol())
                    score -= hs
        # Set as highest if score higher or if score equal and smiles comes first alphabetically
        if not highest or highest['score'] < score or (highest['score'] == score and smiles < highest['smiles']):
            log.debug('New highest tautomer: %s (%s)', smiles, score)
            highest = {'smiles': smiles, 'tautomer': t, 'score': score}
    return highest['tautomer']


def enumerate_tautomers(mol, max_tautomers=1000):
    """Enumerate all possible tautomers and return them as a list.

    :param mol: An RDKit Mol object.
    :param max_tautomers: The maximum number of tautomers to enumerate (limit to prevent combinatorial explosion)
    """
    tautomers = {Chem.MolToSmiles(mol, isomericSmiles=True): copy.deepcopy(mol)}
    done = set()
    while len(tautomers) < max_tautomers:
        for tsmiles in sorted(tautomers):
            if tsmiles in done:
                continue
            for transform in tautomer_transforms:
                for match in tautomers[tsmiles].GetSubstructMatches(transform['smarts']):
                    # Adjust hydrogens
                    product = copy.deepcopy(tautomers[tsmiles])
                    first = product.GetAtomWithIdx(match[0])
                    last = product.GetAtomWithIdx(match[-1])
                    first.SetNumExplicitHs(max(0, first.GetNumExplicitHs() - 1))
                    last.SetNumExplicitHs(last.GetTotalNumHs() + 1)
                    # Adjust bond orders
                    for bi, pair in enumerate(_pairwise(match)):
                        if 'bonds' in transform:
                            product.GetBondBetweenAtoms(*pair).SetBondType(BONDMAP[transform['bonds'][bi]])
                        else:
                            product.GetBondBetweenAtoms(*pair).SetBondType(BondType.DOUBLE if bi % 2 == 0 else BondType.SINGLE)
                    # Adjust charges
                    if 'charges' in transform:
                        for ci, idx in enumerate(match):
                            atom = product.GetAtomWithIdx(idx)
                            atom.SetFormalCharge(atom.GetFormalCharge() + CHARGEMAP[transform['charges'][ci]])
                    try:
                        Chem.SanitizeMol(product)
                        smiles = Chem.MolToSmiles(product, isomericSmiles=True)
                        log.debug('Applied rule: %s to %s', transform['name'], tsmiles)
                        if smiles not in tautomers:
                            log.debug('New tautomer produced: %s' % smiles)
                            tautomers[smiles] = product
                        else:
                            log.debug('Previous tautomer produced again: %s' % smiles)
                    except ValueError:
                        log.debug('ValueError')
            done.add(tsmiles)
        if len(tautomers) == len(done):
            break
    else:
        log.warn('Tautomer enumeration stopped at maximum %s', max_tautomers)
    # Clean up stereochemistry
    for tautomer in tautomers.values():
        Chem.AssignStereochemistry(tautomer, force=True, cleanIt=True)
        for bond in tautomer.GetBonds():
            if bond.GetBondType() == BondType.DOUBLE and bond.GetStereo() > BondStereo.STEREOANY:
                begin = bond.GetBeginAtomIdx()
                end = bond.GetEndAtomIdx()
                for othertautomer in tautomers.values():
                    if not othertautomer.GetBondBetweenAtoms(begin, end).GetBondType() == BondType.DOUBLE:
                        neighbours = tautomer.GetAtomWithIdx(begin).GetBonds() + tautomer.GetAtomWithIdx(end).GetBonds()
                        for otherbond in neighbours:
                            if otherbond.GetBondDir() in {BondDir.ENDUPRIGHT, BondDir.ENDDOWNRIGHT}:
                                otherbond.SetBondDir(BondDir.NONE)
                        Chem.AssignStereochemistry(tautomer, force=True, cleanIt=True)
                        log.debug('Removed stereochemistry from unfixed double bond')
                        break
    return tautomers.values()


def _pairwise(iterable):
    """Utility function to iterate in a pairwise fashion."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


tautomer_enumeration_tests = [
    ('1,3 keto/enol tautomer', 'C1(=CCCCC1)O', {'OC1=CCCCC1', 'O=C1CCCCC1'}),
    ('1,3 keto/enol tautomer', 'C1(CCCCC1)=O', {'OC1=CCCCC1', 'O=C1CCCCC1'}),
    ('Acetophenone keto/enol tautomer', 'C(=C)(O)C1=CC=CC=C1', {'C=C(O)c1ccccc1', 'CC(=O)c1ccccc1'}),
    ('Acetone keto/enol tautomer', 'CC(C)=O', {'CC(C)=O', 'C=C(C)O'}),
    ('keto/enol tautomer', 'OC(C)=C(C)C', {'C=C(O)C(C)C', 'CC(C)=C(C)O', 'CC(=O)C(C)C'}),
    ('1-phenyl-2-propanone enol/keto', 'c1(ccccc1)CC(=O)C', {'C=C(O)Cc1ccccc1', 'CC(=O)Cc1ccccc1', 'CC(O)=Cc1ccccc1'}),
    ('1,5 keto/enol tautomer', 'Oc1nccc2cc[nH]c(=N)c12', {'Nc1nccc2ccnc(O)c12', 'N=c1nccc2cc[nH]c(O)c1-2', 'Nc1[nH]ccc2ccnc(=O)c1-2', 'N=c1[nH]ccc2ccnc(O)c21', 'Nc1nccc2cc[nH]c(=O)c12', 'N=c1[nH]ccc2cc[nH]c(=O)c21'}),
    ('1,5 keto/enol tautomer', 'C1(C=CCCC1)=O', {'O=C1C=CCCC1', 'OC1=CCC=CC1', 'OC1=CC=CCC1', 'O=C1CC=CCC1', 'OC1=CCCC=C1'}),
    ('1,5 keto/enol tautomer', 'C1(=CC=CCC1)O', {'O=C1C=CCCC1', 'OC1=CCC=CC1', 'OC1=CC=CCC1', 'O=C1CC=CCC1', 'OC1=CCCC=C1'}),
    ('aliphatic imine tautomer', 'C1(CCCCC1)=N', {'N=C1CCCCC1', 'NC1=CCCCC1'}),
    ('aliphatic imine tautomer', 'C1(=CCCCC1)N', {'N=C1CCCCC1', 'NC1=CCCCC1'}),
    ('special imine tautomer', 'C1(C=CC=CN1)=CC', {'CC=C1C=CC=CN1', 'CCc1ccccn1', 'CC=C1C=CCC=N1'}),
    ('special imine tautomer', 'C1(=NC=CC=C1)CC', {'CC=C1C=CC=CN1', 'CCc1ccccn1', 'CC=C1C=CCC=N1'}),
    ('1,3 aromatic heteroatom H shift', 'O=c1cccc[nH]1', {'Oc1ccccn1', 'O=c1cccc[nH]1'}),
    ('1,3 aromatic heteroatom H shift', 'Oc1ccccn1', {'Oc1ccccn1', 'O=c1cccc[nH]1'}),
    ('1,3 aromatic heteroatom H shift', 'Oc1ncc[nH]1', {'Oc1ncc[nH]1', 'O=c1[nH]cc[nH]1'}),
    ('1,3 heteroatom H shift', 'OC(C)=NC', {'CN=C(C)O', 'CNC(C)=O', 'C=C(O)NC'}),
    ('1,3 heteroatom H shift', 'CNC(C)=O', {'CN=C(C)O', 'CNC(C)=O', 'C=C(O)NC'}),
    ('1,3 heteroatom H shift', 'S=C(N)N', {'N=C(N)S', 'NC(N)=S'}),
    ('1,3 heteroatom H shift', 'SC(N)=N', {'N=C(N)S', 'NC(N)=S'}),
    ('1,3 heteroatom H shift', 'N=c1[nH]ccn(C)1', {'Cn1ccnc1N', 'Cn1cc[nH]c1=N'}),
    ('1,3 heteroatom H shift', 'CN=c1[nH]cncc1', {'CN=c1ccnc[nH]1', 'CNc1ccncn1', 'CN=c1cc[nH]cn1'}),
    ('1,5 aromatic heteroatom H shift', 'Oc1cccc2ccncc12', {'O=c1cccc2cc[nH]cc1-2', 'Oc1cccc2ccncc12'}),
    ('1,5 aromatic heteroatom H shift', 'O=c1cccc2cc[nH]cc1-2', {'O=c1cccc2cc[nH]cc1-2', 'Oc1cccc2ccncc12'}),
    ('1,5 aromatic heteroatom H shift', 'Cc1n[nH]c2ncnn12', {'C=C1NNc2ncnn21', 'Cc1n[nH]c2ncnn12', 'Cc1nnc2[nH]cnn12', 'C=C1NN=C2N=CNN12', 'Cc1nnc2nc[nH]n12', 'C=C1NN=C2NC=NN12'}),
    ('1,5 aromatic heteroatom H shift', 'Cc1nnc2nc[nH]n12', {'C=C1NNc2ncnn21', 'Cc1n[nH]c2ncnn12', 'Cc1nnc2[nH]cnn12', 'C=C1NN=C2N=CNN12', 'Cc1nnc2nc[nH]n12', 'C=C1NN=C2NC=NN12'}),
    ('1,5 aromatic heteroatom H shift', 'Oc1ccncc1', {'Oc1ccncc1', 'O=c1cc[nH]cc1'}),
    ('1,5 aromatic heteroatom H shift', 'Oc1c(cccc3)c3nc2ccncc12', {'Oc1c2ccccc2nc2ccncc12', 'O=c1c2ccccc2nc2cc[nH]cc1-2', 'O=c1c2ccccc2[nH]c2ccncc21'}),
    ('1,3 and 1,5 aromatic heteroatom H shift', 'Oc1ncncc1', {'Oc1ccncn1', 'O=c1ccnc[nH]1', 'O=c1cc[nH]cn1'}),
    ('1,5 aromatic heteroatom H shift', 'C2(=C1C(=NC=N1)[NH]C(=N2)N)O', {'N=c1[nH]c2ncnc-2c(O)[nH]1', 'Nc1nc2nc[nH]c2c(O)n1', 'N=c1nc(O)c2nc[nH]c2[nH]1', 'N=c1[nH]c2[nH]cnc2c(=O)[nH]1', 'Nc1nc2ncnc-2c(O)[nH]1', 'N=c1nc2nc[nH]c2c(O)[nH]1', 'N=c1nc(O)c2[nH]cnc2[nH]1', 'Nc1nc(O)c2ncnc-2[nH]1', 'Nc1nc(=O)c2nc[nH]c2[nH]1', 'Nc1nc(=O)c2[nH]cnc2[nH]1', 'Nc1nc2[nH]cnc2c(O)n1', 'N=c1nc2[nH]cnc2c(O)[nH]1', 'Nc1nc2[nH]cnc2c(=O)[nH]1', 'Nc1nc2nc[nH]c2c(=O)[nH]1', 'N=c1[nH]c2nc[nH]c2c(=O)[nH]1'}),
    ('1,5 aromatic heteroatom H shift', 'C2(C1=C([NH]C=N1)[NH]C(=N2)N)=O', {'N=c1[nH]c2ncnc-2c(O)[nH]1', 'Nc1nc2nc[nH]c2c(O)n1', 'N=c1nc(O)c2nc[nH]c2[nH]1', 'N=c1[nH]c2[nH]cnc2c(=O)[nH]1', 'Nc1nc2ncnc-2c(O)[nH]1', 'N=c1nc2nc[nH]c2c(O)[nH]1', 'N=c1nc(O)c2[nH]cnc2[nH]1', 'Nc1nc(O)c2ncnc-2[nH]1', 'Nc1nc(=O)c2nc[nH]c2[nH]1', 'Nc1nc(=O)c2[nH]cnc2[nH]1', 'Nc1nc2[nH]cnc2c(O)n1', 'N=c1nc2[nH]cnc2c(O)[nH]1', 'Nc1nc2[nH]cnc2c(=O)[nH]1', 'Nc1nc2nc[nH]c2c(=O)[nH]1', 'N=c1[nH]c2nc[nH]c2c(=O)[nH]1'}),
    ('1,5 aromatic heteroatom H shift', 'Oc1n(C)ncc1', {'Cn1nccc1O', 'CN1N=CCC1=O', 'Cn1[nH]ccc1=O'}),
    ('1,5 aromatic heteroatom H shift', 'O=c1nc2[nH]ccn2cc1', {'O=c1ccn2cc[nH]c2n1', 'Oc1ccn2ccnc2n1', 'O=c1ccn2ccnc2[nH]1'}),
    ('1,5 aromatic heteroatom H shift', 'N=c1nc[nH]cc1', {'N=c1cc[nH]cn1', 'N=c1ccnc[nH]1', 'Nc1ccncn1'}),
    ('1,5 aromatic heteroatom H shift', 'N=c(c1)ccn2cc[nH]c12', {'N=c1ccn2cc[nH]c2c1', 'Nc1ccn2ccnc2c1'}),
    ('1,5 aromatic heteroatom H shift', 'CN=c1nc[nH]cc1', {'CN=c1ccnc[nH]1', 'CNc1ccncn1', 'CN=c1cc[nH]cn1'}),
    ('1,7 aromatic heteroatom H shift', 'c1ccc2[nH]c(-c3nc4ccccc4[nH]3)nc2c1', {'c1ccc2c(c1)=NC(C1=NC3C=CC=CC3=N1)N=2', 'c1ccc2[nH]c(-c3nc4ccccc4[nH]3)nc2c1', 'c1ccc2[nH]c(C3=NC4C=CC=CC4=N3)nc2c1', 'c1ccc2[nH]c(C3N=c4ccccc4=N3)nc2c1', 'c1ccc2c(c1)=NC(=C1N=C3C=CC=CC3N1)N=2', 'c1ccc2c(c1)NC(=C1N=c3ccccc3=N1)N2'}),
    ('1,7 aromatic heteroatom H shift', 'c1ccc2c(c1)NC(=C1N=c3ccccc3=N1)N2', {'c1ccc2c(c1)=NC(C1=NC3C=CC=CC3=N1)N=2', 'c1ccc2[nH]c(-c3nc4ccccc4[nH]3)nc2c1', 'c1ccc2[nH]c(C3=NC4C=CC=CC4=N3)nc2c1', 'c1ccc2[nH]c(C3N=c4ccccc4=N3)nc2c1', 'c1ccc2c(c1)=NC(=C1N=C3C=CC=CC3N1)N=2', 'c1ccc2c(c1)NC(=C1N=c3ccccc3=N1)N2'}),
    ('1,9 aromatic heteroatom H shift', 'CNc1ccnc2ncnn21', {'CN=c1cc[nH]c2ncnn21', 'CN=c1ccnc2[nH]cnn21', 'CN=c1ccnc2nc[nH]n21', 'CNc1ccnc2ncnn21'}),
    ('1,9 aromatic heteroatom H shift', 'CN=c1ccnc2nc[nH]n21', {'CN=c1cc[nH]c2ncnn21', 'CN=c1ccnc2[nH]cnn21', 'CN=c1ccnc2nc[nH]n21', 'CNc1ccnc2ncnn21'}),
    ('1,11 aromatic heteroatom H shift', 'Nc1ccc(C=C2C=CC(=O)C=C2)cc1', {'Nc1ccc(C=C2C=CC(=O)C=C2)cc1', 'N=C1C=CC(=CC2C=CC(=O)C=C2)C=C1', 'N=C1C=CC(=Cc2ccc(O)cc2)C=C1', 'N=C1C=CC(C=C2C=CC(=O)C=C2)C=C1'}),
    ('1,11 aromatic heteroatom H shift', 'N=C1C=CC(=Cc2ccc(O)cc2)C=C1', {'Nc1ccc(C=C2C=CC(=O)C=C2)cc1', 'N=C1C=CC(=CC2C=CC(=O)C=C2)C=C1', 'N=C1C=CC(=Cc2ccc(O)cc2)C=C1', 'N=C1C=CC(C=C2C=CC(=O)C=C2)C=C1'}),
    ('heterocyclic tautomer', 'n1ccc2ccc[nH]c12', {'c1cc2cccnc2[nH]1', 'c1cc2ccc[nH]c-2n1'}),
    ('heterocyclic tautomer', 'c1cc(=O)[nH]c2nccn12', {'O=c1ccn2cc[nH]c2n1', 'Oc1ccn2ccnc2n1', 'O=c1ccn2ccnc2[nH]1'}),
    ('heterocyclic tautomer', 'c1cnc2c[nH]ccc12', {'c1cc2cc[nH]cc-2n1', 'c1cc2ccncc2[nH]1'}),
    ('heterocyclic tautomer', 'n1ccc2c[nH]ccc12', {'c1cc2cnccc2[nH]1', 'c1cc2c[nH]ccc-2n1'}),
    ('heterocyclic tautomer', 'c1cnc2ccc[nH]c12', {'c1cc2[nH]cccc-2n1', 'c1cc2ncccc2[nH]1'}),
    ('furanone tautomer', 'C1=CC=C(O1)O', {'Oc1ccco1', 'O=C1CC=CO1'}),
    ('furanone tautomer', 'O=C1CC=CO1', {'Oc1ccco1', 'O=C1CC=CO1'}),
    ('keten/ynol tautomer', 'CC=C=O', {'CC=C=O', 'CC#CO'}),
    ('keten/ynol tautomer', 'CC#CO', {'CC=C=O', 'CC#CO'}),
    ('ionic nitro/aci-nitro tautomer', 'C([N+](=O)[O-])C', {'CC[N+](=O)[O-]', 'CC=[N+]([O-])O'}),
    ('ionic nitro/aci-nitro tautomer', 'C(=[N+](O)[O-])C', {'CC[N+](=O)[O-]', 'CC=[N+]([O-])O'}),
    ('oxim nitroso tautomer', 'CC(C)=NO', {'CC(C)N=O', 'CC(C)=NO', 'C=C(C)NO'}),
    ('oxim nitroso tautomer', 'CC(C)N=O', {'CC(C)N=O', 'CC(C)=NO', 'C=C(C)NO'}),
    ('oxim/nitroso tautomer via phenol', 'O=Nc1ccc(O)cc1', {'O=NC1C=CC(=O)C=C1', 'O=C1C=CC(=NO)C=C1', 'O=Nc1ccc(O)cc1'}),
    ('oxim/nitroso tautomer via phenol', 'O=C1C=CC(=NO)C=C1', {'O=NC1C=CC(=O)C=C1', 'O=C1C=CC(=NO)C=C1', 'O=Nc1ccc(O)cc1'}),
    ('cyano/iso-cyanic acid tautomer', 'C(#N)O', {'N#CO', 'N=C=O'}),
    ('cyano/iso-cyanic acid tautomer', 'C(=N)=O', {'N#CO', 'N=C=O'}),
    ('isocyanide tautomer', 'C#N', {'[C-]#[NH+]', 'C#N'}),
    ('isocyanide tautomer', '[C-]#[NH+]', {'[C-]#[NH+]', 'C#N'}),
    ('Remove stereochemistry from mobile double bonds', 'c1(ccccc1)/C=C(/O)\\C', {'C=C(O)Cc1ccccc1', 'CC(O)=Cc1ccccc1', 'CC(=O)Cc1ccccc1'}),
    ('Remove stereochemistry from mobile double bonds', 'C/C=C/C(C)=O', {'C=C(O)C=CC', 'C=CCC(=C)O', 'CC=CC(C)=O', 'C=CCC(C)=O', 'C=CC=C(C)O'}),
    ('Remove stereochemistry from mobile double bonds', 'C/C=C\C(C)=O', {'C=C(O)C=CC', 'C=CCC(=C)O', 'CC=CC(C)=O', 'C=CCC(C)=O', 'C=CC=C(C)O'}),
]

tautomer_canonicalization_tests = [
    ('1,3 keto/enol tautomer', 'C1(=CCCCC1)O', 'O=C1CCCCC1'),
    ('1,3 keto/enol tautomer', 'C1(CCCCC1)=O', 'O=C1CCCCC1'),
    ('Acetophenone keto/enol tautomer', 'C(=C)(O)C1=CC=CC=C1', 'CC(=O)c1ccccc1'),
    ('Acetone keto/enol tautomer', 'CC(C)=O', 'CC(C)=O'),
    ('keto/enol tautomer', 'OC(C)=C(C)C', 'CC(=O)C(C)C'),
    ('1-phenyl-2-propanone enol/keto', 'c1(ccccc1)CC(=O)C', 'CC(=O)Cc1ccccc1'),
    ('1,5 keto/enol tautomer', 'Oc1nccc2cc[nH]c(=N)c12', 'N=c1[nH]ccc2cc[nH]c(=O)c21'),
    ('1,5 keto/enol tautomer', 'C1(C=CCCC1)=O', 'O=C1C=CCCC1'),
    ('1,5 keto/enol tautomer', 'C1(=CC=CCC1)O', 'O=C1C=CCCC1'),
    ('aliphatic imine tautomer', 'C1(CCCCC1)=N', 'N=C1CCCCC1'),
    ('aliphatic imine tautomer', 'C1(=CCCCC1)N', 'N=C1CCCCC1'),
    ('special imine tautomer', 'C1(C=CC=CN1)=CC', 'CCc1ccccn1'),
    ('special imine tautomer', 'C1(=NC=CC=C1)CC', 'CCc1ccccn1'),
    ('1,3 aromatic heteroatom H shift', 'O=c1cccc[nH]1', 'O=c1cccc[nH]1'),
    ('1,3 aromatic heteroatom H shift', 'Oc1ccccn1', 'O=c1cccc[nH]1'),
    ('1,3 aromatic heteroatom H shift', 'Oc1ncc[nH]1', 'O=c1[nH]cc[nH]1'),
    ('1,3 heteroatom H shift', 'OC(C)=NC', 'CNC(C)=O'),
    ('1,3 heteroatom H shift', 'CNC(C)=O', 'CNC(C)=O'),
    ('1,3 heteroatom H shift', 'S=C(N)N', 'NC(N)=S'),
    ('1,3 heteroatom H shift', 'SC(N)=N', 'NC(N)=S'),
    ('1,3 heteroatom H shift', 'N=c1[nH]ccn(C)1', 'Cn1cc[nH]c1=N'),
    ('1,3 heteroatom H shift', 'CN=c1[nH]cncc1', 'CN=c1cc[nH]cn1'),
    ('1,5 aromatic heteroatom H shift', 'Oc1cccc2ccncc12', 'Oc1cccc2ccncc12'),
    ('1,5 aromatic heteroatom H shift', 'O=c1cccc2cc[nH]cc1-2', 'Oc1cccc2ccncc12'),
    ('1,5 aromatic heteroatom H shift', 'Cc1n[nH]c2ncnn12', 'Cc1n[nH]c2ncnn12'),
    ('1,5 aromatic heteroatom H shift', 'Cc1nnc2nc[nH]n12', 'Cc1n[nH]c2ncnn12'),
    ('1,5 aromatic heteroatom H shift', 'Oc1ccncc1', 'O=c1cc[nH]cc1'),
    ('1,5 aromatic heteroatom H shift', 'Oc1c(cccc3)c3nc2ccncc12', 'O=c1c2ccccc2[nH]c2ccncc21'),
    ('1,3 and 1,5 aromatic heteroatom H shift', 'Oc1ncncc1', 'O=c1cc[nH]cn1'),
    ('1,5 aromatic heteroatom H shift', 'C2(=C1C(=NC=N1)[NH]C(=N2)N)O', 'N=c1[nH]c2[nH]cnc2c(=O)[nH]1'),
    ('1,5 aromatic heteroatom H shift', 'C2(C1=C([NH]C=N1)[NH]C(=N2)N)=O', 'N=c1[nH]c2[nH]cnc2c(=O)[nH]1'),
    ('1,5 aromatic heteroatom H shift', 'Oc1n(C)ncc1', 'Cn1[nH]ccc1=O'),
    ('1,5 aromatic heteroatom H shift', 'O=c1nc2[nH]ccn2cc1', 'O=c1ccn2cc[nH]c2n1'),
    ('1,5 aromatic heteroatom H shift', 'N=c1nc[nH]cc1', 'N=c1cc[nH]cn1'),
    ('1,5 aromatic heteroatom H shift', 'N=c(c1)ccn2cc[nH]c12', 'N=c1ccn2cc[nH]c2c1'),
    ('1,5 aromatic heteroatom H shift', 'CN=c1nc[nH]cc1', 'CN=c1cc[nH]cn1'),
    ('1,7 aromatic heteroatom H shift', 'c1ccc2[nH]c(-c3nc4ccccc4[nH]3)nc2c1', 'c1ccc2[nH]c(-c3nc4ccccc4[nH]3)nc2c1'),
    ('1,7 aromatic heteroatom H shift', 'c1ccc2c(c1)NC(=C1N=c3ccccc3=N1)N2', 'c1ccc2[nH]c(-c3nc4ccccc4[nH]3)nc2c1'),
    ('1,9 aromatic heteroatom H shift', 'CNc1ccnc2ncnn21', 'CN=c1cc[nH]c2ncnn21'),
    ('1,9 aromatic heteroatom H shift', 'CN=c1ccnc2nc[nH]n21', 'CN=c1cc[nH]c2ncnn21'),
    ('1,11 aromatic heteroatom H shift', 'Nc1ccc(C=C2C=CC(=O)C=C2)cc1', 'Nc1ccc(C=C2C=CC(=O)C=C2)cc1'),
    ('1,11 aromatic heteroatom H shift', 'N=C1C=CC(=Cc2ccc(O)cc2)C=C1', 'Nc1ccc(C=C2C=CC(=O)C=C2)cc1'),
    ('heterocyclic tautomer', 'n1ccc2ccc[nH]c12', 'c1cc2cccnc2[nH]1'),
    ('heterocyclic tautomer', 'c1cc(=O)[nH]c2nccn12', 'O=c1ccn2cc[nH]c2n1'),
    ('heterocyclic tautomer', 'c1cnc2c[nH]ccc12', 'c1cc2ccncc2[nH]1'),
    ('heterocyclic tautomer', 'n1ccc2c[nH]ccc12', 'c1cc2cnccc2[nH]1'),
    ('heterocyclic tautomer', 'c1cnc2ccc[nH]c12', 'c1cc2ncccc2[nH]1'),
    ('furanone tautomer', 'C1=CC=C(O1)O', 'Oc1ccco1'),
    ('furanone tautomer', 'O=C1CC=CO1', 'Oc1ccco1'),
    ('keten/ynol tautomer', 'CC=C=O', 'CC=C=O'),
    ('keten/ynol tautomer', 'CC#CO', 'CC=C=O'),
    ('ionic nitro/aci-nitro tautomer', 'C([N+](=O)[O-])C', 'CC[N+](=O)[O-]'),
    ('ionic nitro/aci-nitro tautomer', 'C(=[N+](O)[O-])C', 'CC[N+](=O)[O-]'),
    ('oxim nitroso tautomer', 'CC(C)=NO', 'CC(C)=NO'),
    ('oxim nitroso tautomer', 'CC(C)N=O', 'CC(C)=NO'),
    ('oxim/nitroso tautomer via phenol', 'O=Nc1ccc(O)cc1', 'O=Nc1ccc(O)cc1'),
    ('oxim/nitroso tautomer via phenol', 'O=C1C=CC(=NO)C=C1', 'O=Nc1ccc(O)cc1'),
    ('cyano/iso-cyanic acid tautomer', 'C(#N)O', 'N=C=O'),
    ('cyano/iso-cyanic acid tautomer', 'C(=N)=O', 'N=C=O'),
    ('formamidinesulfinic acid tautomer', '[S](=O)(=O)C(N)N', 'N=C(N)S(=O)O'),
    ('formamidinesulfinic acid tautomer', '[S](=O)(O)C(=N)N', 'N=C(N)S(=O)O'),
    ('isocyanide tautomer', 'C#N', 'C#N'),
    ('isocyanide tautomer', '[C-]#[NH+]', 'C#N'),
]

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    for desc, smiles, tautomers in tautomer_enumeration_tests:
        mol = Chem.MolFromSmiles(smiles)
        Chem.SanitizeMol(mol)
        assert {Chem.MolToSmiles(t, isomericSmiles=True) for t in enumerate_tautomers(mol)} == tautomers
        log.info('%s => %s', smiles, tautomers)

    for desc, smiles, tautomer in tautomer_canonicalization_tests:
        mol = Chem.MolFromSmiles(smiles)
        Chem.SanitizeMol(mol)
        assert Chem.MolToSmiles(canonical_tautomer(mol), isomericSmiles=True) == tautomer
        log.info('%s => %s', smiles, tautomer)




# get-ebi.py
# Copyright 2005, 2006 by Bernhard Lohkamp
# Copyright 2005, 2006 by Paul Emsley, The University of York
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import numbers
import coot
import coot_utils
import pdbe_validation_data

# was
# https://www.ebi.ac.uk/pdbe-srv/view/files/1sar.ent
# now
# https://www.ebi.ac.uk/pdbe/entry-files/download/pdb1sar.ent

pdbe_server = "https://www.ebi.ac.uk"
# 20230501-PE which is it? (merge conflict - keep both)
pdbe_pdb_file_dir = "pdbe-srv/view/files"
pdbe_pdb_file_dir = "pdbe/entry-files/download"

pdbe_file_name_tail = "ent"

# sf example http://www.ebi.ac.uk/pdbe-srv/view/files/r4hrhsf.ent

# 20151126-PE No, we can't have coot-download created on coot-startup, it must be
#             made only when we need it.
# global coot_tmp_dir
# coot_tmp_dir = coot_utils.get_directory("coot-download")

# e.g. (ebi-get-pdb "1crn")
# 
# no useful return value
#
# Note that that for sf data, we need to construct something like the
# string: http://oca.ebi.ac.uk/oca-bin/send-sf?r2acesf.ent.Z and we
# don't need to strip any html (thank goodness). Also not that the
# accession code now is lower case.
#
# data-type(here string) can be 'pdb' or 'sfs' (structure factors). 
# We might like to use
# 'coordinates rather than 'pdb in the future.
# 
# The optional argument imol-coords-arg-list is necessary for
# outputting sfs, because we need coordinates from which we can
# calculate phases.
#

# helper function to avoid downloading empty files
# returns download filename upon success or False when fail
# should we return the response header too?
#
def coot_urlretrieve(url, file_name, reporthook=None):

    """Helper function to avoid downloading empty files
    returns download filename upon success or False when fail."""

    import urllib.request, urllib.parse, urllib.error
    local_filename = False
    class CootURLopener(urllib.request.FancyURLopener):
        def http_error_default(self, url, fp, errcode, errmsg, headers):
            # handle errors the way you'd like to
            # we just pass
            pass

    # opener = CootURLopener(context=ssl_context)
    opener = CootURLopener()
    try:
        local_filename, header = opener.retrieve(url, file_name, reporthook)
    except:
        # we could catch more here, but dont bother for now
        print("WARNING:: retrieve of url %s failed" %url)

    return local_filename

# we dont need something like net-get-url in python 
# since we have build in functions like urlretrieve (in module urllib)

# check the directory and get url url_string.
#
def check_dir_and_get_url(dir, file_name, url_string):
    import os,urllib.request,urllib.parse,urllib.error

    # FIXME logic, can be done better
    if (os.path.isfile(dir) or os.path.isdir(dir)):
       if (os.path.isfile(dir)):
          print(dir, " is atually a file and not a dir, so we can't write to it")
       else:
          if (os.path.isdir(dir)):
              coot_urlretrieve(url_string, file_name)
          else:
              print("ERROR:: Oops - Can't write to ", dir, " directory!")
    else:
       os.makedirs(dir)
       if (os.path.isdir(dir)):
           coot_urlretrieve(url_string, file_name)
       else:
         print("ERROR:: Oops - create-directory ",dir," failed!")

# get url_string for data type (string actually) 'pdb' or 'sfs'
#
def get_url_str(id, url_string, data_type, imol_coords_arg_list):
    import operator

    #print "DEBUG:: in get_url_string:", id, url_string, data_type

    coot_download_dir = coot_utils.get_directory("coot-download")
    if data_type == "pdb":
       pdb_file_name = coot_download_dir + "/" + id + ".pdb." + \
           pdbe_file_name_tail
       check_dir_and_get_url(coot_download_dir, pdb_file_name, url_string)
       imol_coords = coot.handle_read_draw_molecule(pdb_file_name)
       return imol_coords

    if data_type == "cif":
       pdb_file_name = coot_download_dir + "/" + id + ".cif"
       check_dir_and_get_url(coot_download_dir, pdb_file_name, url_string)
       imol_coords = coot.handle_read_draw_molecule(pdb_file_name)
       return imol_coords

    if data_type == "sfs":
       sfs_file_name = coot_download_dir + "/" + id + ".cif"
       #       print "BL DEBUG:: cif output file is: ",sfs_file_name
       imol_coords = imol_coords_arg_list
       if (isinstance(imol_coords, numbers.Number) and imol_coords>=-1):
         check_dir_and_get_url(coot_tmp_dir, sfs_file_name, url_string)
         coot.read_cif_data(sfs_file_name, imol_coords_arg_list)
         # do we need to return something here too?!

# Get the pdb and sfs. @var{id} is the accession code
#
def get_ebi_pdb_and_sfs(id):

    import operator,string

    imol_coords = get_ebi_pdb(id)
    if (not isinstance(imol_coords, numbers.Number)):
       print("Failed at reading coordinates. imol-coords was ", imol_coords)

    if (imol_coords < 0):	# -1 is coot code for failed read.
       print("failed to read coordinates.")
    else:
       down_id = id.lower()
       url_str = pdbe_server + "/" + pdbe_pdb_file_dir + "/" + \
                 "r" + down_id + "sf." + \
                 pdbe_file_name_tail
       get_url_str(id, url_str, "sfs", imol_coords)

# Return a molecule number on success
# or not a number (False) or -1 on error.
#
def get_ebi_pdb(id):
    import urllib.request, urllib.parse, urllib.error, string

    # print "======= id:", id
    down_id = id.lower()
    pdb_url_str = pdbe_server + "/" + pdbe_pdb_file_dir + "/" + down_id + ".ent"
    cif_url_str = pdbe_server + "/" + pdbe_pdb_file_dir + "/" + down_id + ".cif"
    url_status = get_url_str(id, pdb_url_str, "pdb", None)
    # e.g. http://ftp.ebi.ac.uk/pub/databases/pdb +
    #      /validation_reports/cb/1cbs/1cbs_validation.xml.gz
    # print "BL DEBUG:: get-ebi-pdb ======= url-status", url_status
    if coot_utils.valid_model_molecule_qm(url_status):
        pdbe_validation_data.pdb_validate(down_id, url_status)
        return url_status
    else:
        cif_url_status = get_url_str(id, cif_url_str, "cif", None)
        if coot_utils.valid_model_molecule_qm(cif_url_status):
            # 2023-06-12-PE remove this for now - it doesn't work with GTK4.
            # pdbe_validation_data.pdb_validate(down_id, cif_url_status)
            return cif_url_status

    return False


# Return a list of molecules (i.e. the model molecule and the 2 maps).
# or, if it didn't work then return False
#
# @var{id} is the accession code.
#
# returns imol of read pdb or False on error.
#
# 20050725 EDS code
#
# return a list of 3 molecule numbers [imol, map, diff_map] or False
#
#
def get_eds_pdb_and_mtz(id):
    import string
    import urllib.request, urllib.parse, urllib.error

    # Gerard DVD Kleywegt says we can find the coords/mtz thusly:
    #
    # - model = http://eds.bmc.uu.se/eds/sfd/1cbs/pdb1cbs.ent
    # - mtz   = http://eds.bmc.uu.se/eds/sfd/1cbs/1cbs_sigmaa.mtz
    #
    # 20091222
    # - newprefix: http://eds.bmc.uu.se/eds/dfs/cb/1cbs/
    # 
    # URL:: "http://eds.bmc.uu.se/eds/sfd/sa/2sar/pdb2sar.ent"
    # URL:: "http://eds.bmc.uu.se/eds/sfd/sa/2sar/2sar_sigmaa.mtz"
    #
    # 20161010 new prefix
    # http://www.ebi.ac.uk/pdbe/coordinates/
    # http://www.ebi.ac.uk/pdbe/coordinates/files/1cbs_map.mtz

    def get_cached_eds_files(accession_code):
        down_code = accession_code.lower()
        dir_name = coot_utils.get_directory("coot-download")
        pdb_file_name = os.path.join(dir_name, "pdb" + down_code + ".ent")
        mtz_file_name = os.path.join(dir_name, down_code + "_map.mtz")

        print( "::::::::: pdb_file_name:", pdb_file_name)
        print( "::::::::: mtz_file_name:", mtz_file_name)
        if not os.path.isfile(pdb_file_name):
            return False
        else:
            if not os.path.isfile(mtz_file_name):
                return False
            else:
                imol = coot.read_pdb(pdb_file_name)
                imol_map = coot.make_and_draw_map(mtz_file_name, "FWT", "PHWT", "", 0, 0)
                imol_map_d = coot.make_and_draw_map(mtz_file_name, "DELFWT", "PHDELWT", "", 0, 1)
                if not (coot_utils.valid_model_molecule_qm(imol) and
                        coot_utils.valid_map_molecule_qm(imol_map) and
                        coot_utils.valid_map_molecule_qm(imol_map_d)):
                    coot.close_molecule(imol)
                    coot.close_molecule(imol_map)
                    coot.close_molecule(imol_map_d)
                    return False
                else:
                    return [imol, imol_map, imol_map_d]

    eds_site = "https://www.ebi.ac.uk/pdbe/coordinates"
    # https://www.ebi.ac.uk/pdbe/entry/pdb/6tje
    # eds_core = "some://thing" ;; for web pages
    eds_core = "https://www.ebi.ac.uk/pdbe/entry/pdb"
    eds_coords_site = "https://www.ebi.ac.uk/pdbe/entry-files/download"
    # now the map mtz files are like this:
    # https://www.ebi.ac.uk/pdbe/coordinates/files/zz/4zzn/4zzn_map.mtz

    # "1cbds" -> "cb/"
    #
    def mid_chars(id_code):
        if not id_code:  # check for string?
            return "//fail//"
        if not (len(id_code) == 4):
            return "/FAIL/"
        else:
            return id_code[1:3] + "/"

    # main line
    #
    cached_status = get_cached_eds_files(id)
    if isinstance(cached_status, list):
        return cached_status
    else:
        coot_tmp_dir = coot_utils.get_directory("coot-download")	
        r = coot_utils.coot_mkdir(coot_tmp_dir)

        if r:
            down_id = id.lower()
            target_pdb_file = "pdb" + down_id + ".ent"
            target_cif_file = down_id + ".cif"
            dir_target_pdb_file = coot_tmp_dir + "/" + target_pdb_file
            dir_target_cif_file = coot_tmp_dir + "/" + target_cif_file
            model_url = eds_coords_site + "/" + target_pdb_file
            model_cif_url = eds_coords_site + "/" + target_cif_file
            target_mtz_file = down_id + "_map.mtz"
            dir_target_mtz_file = coot_tmp_dir + "/" + target_mtz_file
            # mtz_url = eds_site  + "/files/" + target_mtz_file
            #mtz_url = eds_site + "/files/" + mid_chars(down_id) + "/" + \
            #          down_id + "/" + down_id + "_map.mtz"
            mtz_url = eds_coords_site + "/" + down_id + "_map.mtz"
            eds_info_page = eds_core + "/" + down_id
            bad_map_status = False

            print("model_url:", model_url)
            print("  mtz_url:", mtz_url)

            # I am not sure this exists now
            # print("eds_info_page:", eds_info_page)

            try:
                # pre_download_info = coot.coot_get_url_as_string(eds_info_page)
                # print "INFO:: --------------- pre-download-info:", pre_download_info
                #  bad_map_status = "No reliable map available" in pre_download_info
                # if "There is no structure factor entry" in pre_download_info:
                #     print("WARNING:: no sfs available for entry %s, so wont download." %id)
                #     # no pdb and no mtz
                #    return False
                pass
            except:
                print("ERROR:: could not get pre_download_info from", eds_core)
                # we probably wont get anything else, so bail out.
                return False

            s1 = coot_urlretrieve(model_url, dir_target_pdb_file)
            s2 = coot_urlretrieve(mtz_url, dir_target_mtz_file)

            if bad_map_status:
                s = "This map (" + down_id + ") is marked by the EDS as \"not a reliable map\""
                coot.info_dialog(s)

            # maybe should then not load the map!?

            print("INFO:: read pdb model status: ",s1)
            print("INFO:: read mtz data  status: ",s2)

            if s1 and os.path.isfile(s1):
                r_imol = coot.handle_read_draw_molecule(dir_target_pdb_file)
                if not coot_utils.valid_model_molecule_qm(r_imol):
                    s1_cif = coot_urlretrieve(model_cif_url, dir_target_cif_file)
                    print("INFO:: read cif model status: ",s1_cif)
                    if (s1_cif == 0):
                        r_imol = coot.handle_read_draw_molecule(dir_target_pdb_file)
                        if not coot_utils.valid_model_molecule_qm(r_imol):
                            return False
                        else:
                            return r_imol
                    else:
                        return False
            if s2 and os.path.isfile(s2):
                map_1 = coot.make_and_draw_map(dir_target_mtz_file, "FWT", "PHWT","",0,0)
                map_2 = coot.make_and_draw_map(dir_target_mtz_file, "DELFWT", "PHDELWT",
                              "", 0, 1)
                coot.set_scrollable_map(map_1)
                return [map_1, map_2]  # r_imol not in this scope (at the moment)
            else:
                return False

        else:
            print("Can't make directory ",coot_tmp_dir)

# not sure if coot function is better or python script function coot_urlretrieve
# return 0 on success
def net_get_url(my_url, file_name):
    coot.coot_get_url(my_url, file_name)

def get_pdb_redo(text):

    if not isinstance(text, str):
        print("WARNING:: No string. No accession code.")
    else:
        if not (len(text) == 4):
            print("WARNING:: Accession code not 4 chars.")
        else:
            text = text.lower()
            stub = "https://pdb-redo.eu/db/" + \
                   text + "/" + text + "_final"
            pdb_file_name = text + "_final.pdb"
            mtz_file_name = text + "_final.mtz"
            py_file_name = text + ".py"
            url_pdb = stub + ".pdb"
            url_mtz = stub + ".mtz"
            url_py = stub + ".py"

            print("DEBUG:: getting", url_pdb)
            status = net_get_url(url_pdb, pdb_file_name)
            if not status == 0:
                print("Failed to get %s %s status %s" %(url_pdb, pdb_file_name, status))
            print("DEBUG:: getting", url_mtz)
            status = net_get_url(url_mtz, mtz_file_name)
            if not status == 0:
                print("Failed to get %s %s status %s" %(url_mtz, mtz_file_name, status))
            print("DEBUG:: getting", url_py)
            status = net_get_url(url_py, py_file_name)
            if not status == 0:
                print("Failed to get %s %s status %s" %(url_py, py_file_name, status))

            status_imol = coot.read_pdb(pdb_file_name)
            if status_imol < 0:
                print("INFO:: problem opening pdb file. Most likely something went wrong in the download")
            else:
                print("DEBUG:: make-and-draw-map with", mtz_file_name)
                coot.make_and_draw_map(mtz_file_name, "FWT", "PHWT", "", 0, 0)
                coot.make_and_draw_map(mtz_file_name, "DELFWT", "PHDELWT", "", 0, 1)
                anom_map = coot.make_and_draw_map(mtz_file_name, "FAN", "PHAN", "", 0, 1)
                if anom_map > -1:
                    coot.set_map_colour(anom_map, 0.5, 0.5, 0)
                exec(compile(open(py_file_name, "rb").read(), py_file_name, 'exec'))
            

# BL says: to test, some examples

#id = "2BSX"
#get_ebi_pdb_and_sfs(id)
#get_eds_pdb_and_mtz(id)



global cache_ligand_metrics
cache_ligand_metrics = False

# return 3 things
#
def mtz_file_name2refinement_column_labels(file_name):

    columns = coot.get_mtz_columns(file_name)
    read_success = columns.read_success

    if (not read_success == 1):

        # unhappy path
        print("Failed to read columms from file %s for map molecule %s" \
              %(file_name, imol_map()))

    else:
        
        # happy path
        f_cols = columns.f_cols
        sigf_cols = columns.sigf_cols
        rfree_cols = columns.r_free_cols

        l1 = f_cols.size()
        l2 = sigf_cols.size()
        l3 = rfree_cols.size()

        if (not all([x > 0 for x in [l1, l2 ,l3]])):

            # unhappy path
            print("Failed to find columns of the necessary types from %s: %s %s %s!" \
                  %(file_name, l1, l2, l3))
            return False, False, False
            
        else:

            # happy path
            f_col_label = f_cols[0].column_label
            sigf_col_label = sigf_cols[0].column_label
            r_free_col_label = rfree_cols[0].column_label

            return f_col_label, sigf_col_label, r_free_col_label


def ligand_validation_metrics_gui_list_wrapper_pre(
    imol, chain_id, res_no, ins_code, refmac_input_mtz_file_name,
    fp_col, sigfp_col, rfree_col, refmac_dir):

    global cache_ligand_metrics

    m = ligand_check.get_metrics_for_ligand(imol, chain_id, res_no, ins_code,
                               refmac_input_mtz_file_name,
                               fp_col, sigfp_col, rfree_col, refmac_dir)

    if cache_ligand_metrics:
        m = [0.935339450836182, 10.6290054321289,
             [0, 10, 18, 70, 143],
             [-0.0306930494488489, 0.386785203839611, 0.0146206432429434, 10865.0,
              5373.01106146351, -27.3189968762454, 0.0173681300165985, 0.0280039555722363,
              -8.86567375069092e-10, -0.0025144037621947, 0.117837198078632, 0.120915851909265],
             [0.952432048573266, 13.3150000572205, 13.9800000190735, 0.176323987538941]]

        if (not isinstance(m, list)):
            # unhappy path
            print("WARNING:: no ligand metrics found.")
        else:
            # happy path ;-)
            diff_d = 0.05
            d = m[0]
            mwz = m[1]           # Mogul Z
            contact_info = m[2]  # number of bad contacts
            bc = contact_info[0]
            low_is_good = 1
            high_id_good = 0

            # if mogul ran OK, then we can display the mogul markup
            #
            if (isinstance(mwz, list)):
                mogul_out_file_name = mwz[1]
                coot.mogul_markup(imol, chain_id, res_no, ins_code,
                             mogul_out_file_name)
                mwz = mwz[0]

            percentile_d = coot.get_ligand_percentile("density_correlation", d, high_id_good)
            percentile_diff_d = coot.get_ligand_percentile("coot_diff_map_correlation", diff_d, low_is_good)
            percentile_mwz = coot.get_ligand_percentile("mogul_z_worst", mwz, low_is_good)
            percentile_bc = coot.get_ligand_percentile("bumps_1", bc, low_is_good)

            if (percentile_d < 0):
                # just an example, but means we do not have ligands-2016.db
                txt = "BL INFO:: we dont have ligands-2016.db, so \n" + \
                      "percentiles are no available and graph meaningless!"
                coot.info_dialog(txt)
            input_to_sliders = [["Direct map density correl.", percentile_d, d],
                                [" Diff map density correl.", percentile_diff_d, diff_d],
                                ["            Mogul Z-worst", percentile_mwz, mwz],
                                ["             Bad contacts", percentile_bc, bc]]

            ligand_validation_sliders.ligand_validation_metrics_gui_list_wrapper(input_to_sliders)
                
            
        
        
#if (use_gui_qm != 2):
# disable for now since I cannot run mogul and test it in real...
if (0):
    menu = coot_gui.coot_menubar_menu("Ligand")

    def ligand_metric_slider_func():
        
        imol_map = coot.imol_refinement_map()

        if (not coot_utils.valid_map_molecule_qm(imol_map)):
            coot.add_status_bar_text("No valid refinement map molecule")
        else:
            l = refmac_parameters(imol_map)
            if l:
                refmac_input_mtz_file_name = l[0]
            else:
                refmac_input_coot.mtz_file_name = mtz_file_name(imol_map)
            refmac_dir = coot_utils.get_directory("coot-refmac")

            f_col_label, sigf_col_label, r_free_col_label = \
                         mtz_file_name2refinement_column_labels(refmac_input_mtz_file_name)
            print("    f_col_label:", f_col_label)
            print(" sigf-col-label:", sigf_col_label)
            print("rfree-col-label:", r_free_col_label)

            with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                       aa_ins_code, aa_atom_name, aa_alt_conf]:
                ligand_validation_sliders.ligand_validation_metrics_gui_list_wrapper_pre(
                    aa_imol, aa_chain_id, aa_res_no, aa_ins_code,
                    refmac_input_mtz_file_name,
                    f_col_label, sigf_col_label, r_free_col_label,
                    refmac_dir)

    add_simple_coot_menu_menuitem(
        menu, "Ligand Metric Sliders",
        lambda func: ligand_metric_slider_func())
        
            
        



#!/usr/bin/env python
"""A class for storing chains of nucleotides."""

# Copyright 2010, 2011 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

from suite import Suite

class Chain:
    """A class for storing chains of nucleotides (which must all be instances of the nucleotide class)."""
    
    def __init__(self, nucs = None):
        """Initialize a Chain object.
        
        OPTIONAL ARGUMENTS:
            nucs - a list of nucleotide objects to add to the chain
                   If not given, then an empty chain is created
        RETURNS:
            an initialized Chain object
        """
        
        #self.__nucNums = []
        self.__resNumDict = {}
        
        if nucs is not None:
            for (curSeqNum, curNuc) in enumerate(nucs):
                curNuc.chain = self
                curNuc.seqNum = curSeqNum
                self.__resNumDict[str(curNuc.resNum)] = curSeqNum
                #self.__nucNums.append(str(curNuc.resNum))
        else:
            nucs = []
        
        self.nucleotides = nucs
    
    @property
    def nucs(self):
        """A synonym for Chain.nucleotides"""
        return self.nucleotides
    
    
    def addNuc(self, nuc):
        """Add a nucleotide to the end of the chain
        
        ARGUMENTS:
            nuc - a nucleotide object to add to the chain
        RETURNS:
            None
        EFFECTS:
            Adds nuc to the chain
        """
        
        nuc.chain = self
        nuc.seqNum = self.numNucs()
        self.nucleotides.append(nuc)
        self.__resNumDict[str(nuc.resNum)] = nuc.seqNum
        #self.__nucNums.append(str(nuc.resNum))
    
    
    def addNuc5p(self, nuc):
        """Add a nucleotide to the start (i.e. 5' end) of the chain
        
        ARGUMENTS:
            nuc - a nucleotide object to add to the chain
        RETURNS:
            None
        EFFECTS:
            Adds nuc to the chain
        """
        
        nuc.chain = self
        nuc.seqNum = 0
        
        #update the sequence number of the nucleotides
        for curNuc in self.nucleotides:
            curNuc.seqNum += 1
        
        #update the residue number lookup dictionary
        for (curKey, curVal) in self.__resNumDict.iteritems():
            self.__resNumDict[curKey] = curVal + 1
        
        self.nucleotides.insert(0, nuc)
    
    
    def suites(self):
        """A generator for suites in the chain
        
        ARGUMENTS:
            None
        YIELDS:
            The next suite of the chain
        """
        
        #if the first nucleotide doesn't have a C1', then assume it's not a complete nucleotide and skip it
        if self.nucleotides[0].hasAtom("C1'"):
            startingNuc = 0
        else:
            startingNuc = 1
        
        for i in xrange(startingNuc, len(self.nucleotides)-2):
            yield Suite(self.nucleotides[i], self.nucleotides[i+1])
    
    
    def len(self):
        """The length of the chain in nucleotides
        
        ARGUMENTS:
            None
        RETURNS:
            The length of the chain in nucleotides
        """
        
        return len(self.nucleotides)
    
    
    def numNucs(self):
        """The length of the chain in nucleotides
        
        ARGUMENTS:
            None
        RETURNS:
            The length of the chain in nucleotides
        """
        
        return len(self.nucleotides)
    
    
    def numSuites(self):
        """The length of the chain in suites
        
        ARGUMENTS:
            None
        RETURNS:
            The length of the chain in suites
        """
        
        return len(self.nucleotides) - 1
    
    
    def resIndex(self, resNum):
        """Find the index (i.e. position in the chain) of a given residue number
        
        ARGUMENTS:
            resNum - the residue number to find (which may include an insertion code)
        RETURNS:
            the index of the specified residue number in this chain
        """
        
        return self.__resNumDict[str(resNum)]
    
    
    def resNumList(self):
        return [nuc.resNum for nuc in self.nucleotides]
        
    def resTypeList(self):
        return [nuc.type for nuc in self.nucleotides]
    
    def __str__(self):
        """Create a string representation of the object"""
        return "\n\n".join(map(str, self.nucleotides))



# a whole set of amercanisms
 
# an americanism
# deftexi set_rotation_center
set_rotation_center =  set_rotation_centre

# an americanism
# deftexi handle_read_draw_molecule_with_recenter
handle_read_draw_molecule_with_recenter = handle_read_draw_molecule_with_recentre

# an americanism
# deftexi set_rotation_center_size 
set_rotation_center_size = set_rotation_centre_size

# an americanism
# deftexi center_atom_label_status
center_atom_label_status = centre_atom_label_status

# an americanism
# deftexi set_last_map_color
set_last_map_color = set_last_map_colour

# an americanism
# deftexi center_of_mass
center_of_mass = centre_of_mass

# an americanism
# set deftexi set_swap_difference_map_colors
set_swap_difference_map_colors = set_swap_difference_map_colours

# an americanism
# deftexi set_font_color
set_font_color = set_font_colour

# an americanism
# deftexi set_dots_color
set_dots_color = set_dots_colour

# an americanism
invert_chiral_center = invert_chiral_centre


# some more from BL

# an americanism
# deftexi rotation_center
rotation_center = rotation_centre

# an americanism
# deftexi move_molecule_to_screen_center
move_molecule_to_screen_center = move_molecule_to_screen_centre
